// osgFX - Copyright (C) 2003 Marco Jez

// AMD /////////////////////////////////////////////////////////////////
// DESCRIPTION: This code is looslely based on the osgFX Effect class. 
// Render target support has been added.
// Copyright (c) 2008 Advanced Micro Devices, Inc. 
////////////////////////////////////////////////////////////////////////

#ifndef __COLLADA_EFFECT_NODE__
#define __COLLADA_EFFECT_NODE__

#include <vector>

#include <osg/Material>
#include <osg/LineWidth>
#include <osg/GL>
#include <osg/PolygonOffset>
#include <osg/PolygonMode>
#include <osg/buffered_value>
#include <osg/ref_ptr>
#include <osg/Node>
#include <osg/Group>
#include <osg/Geode>
#include <osg/OccluderNode>

#include <dom/domProfile_COMMON.h>

#include "ColladaRTT.h"
#include "ColladaTechnique.h"

// Default time base length used by RenderMonkey. This can be defined as an extra in the Effect element
// and is output by RenderMonkey
const float fDefaultRenderMonkeyTimeCycle = 120.0f; 

class daeElement;

class  ColladaEffectNode:  public osg::Group
{
public:

   ColladaEffectNode();
   
   ~ColladaEffectNode();

   ColladaEffectNode::ColladaEffectNode(const ColladaEffectNode& copy, const osg::CopyOp& copyop);

   virtual inline bool isSameKindAs(const osg::Object* obj) const { return dynamic_cast<const ColladaEffectNode*>(obj) != NULL; }
   virtual inline const char* libraryName() const { return "Collada"; }
   virtual inline const char* className() const { return "ColladaEffect"; }

   // AMD added
   ColladaRTT * AddRenderTarget (   const std::string    &  strName, 
                                    int                     nTextureIndex );
   // AMD added
   void ClearRenderTargets ( )
   {
      m_RTTNameMap.clear();
      m_RTTIndexMap.clear();
   }

   // AMD added
   ColladaRTT * GetRenderTargetByName ( std::string strName );

   // AMD added
   ColladaRTT * GetRenderTargetByIndex ( unsigned int nIndex );

   // AMD added
   void AddGeometry ( osg::Node * pGeom ) ;

   // AMD added
   virtual void HandleUpdate ( osg::NodeVisitor& nv );

   // AMD added
   void SetTimeCycle ( float fTimeCycle ) 
   {
      m_fTimeCycle = fTimeCycle ;
   }

   // AMD added
   float GetTimeCycle () 
   {
      return m_fTimeCycle;
   }


   /**
   optional: set effect parameters to produce a visually significant
   result to be used in demo applications like osgfxbrowser. Default
   is to do nothing.
   */
   inline virtual void setUpDemo()   {}

   /** get the number of techniques defined for this Effect */
   inline int getNumTechniques() const;

   /** get the i-th Technique */
   inline ColladaTechnique* getTechnique(int i);

   /** get the i-th const Technique */
   inline const ColladaTechnique* getTechnique(int i) const;

   /** get the index of the currently selected Technique */
   inline int getSelectedTechnique() const;

   enum TechniqueSelection 
   {
      AUTO_DETECT = -1
   };

   /** select a technique or enable automatic detection */
   inline void selectTechnique(int i = AUTO_DETECT);

   /** custom traversal */
   virtual void traverse(osg::NodeVisitor& nv);

   /** default traversal */
   // AMD added
   void inherited_traverse( osg::NodeVisitor& nv, int nPassIndex = -1 );

   /** add a technique to the Effect */
   inline void addTechnique( ColladaTechnique* tech );

protected:

   // AMD added
   void PrintMatrix ( osg::Matrixf & mat );

   // AMD added
   ColladaEffectNode& operator=(const ColladaEffectNode&) { return *this; }

   // AMD added
   osg::MatrixList GetWorldMatrices( osg::Node* haltTraversalAtNode, bool bIgnoreCameras ) const;

private:

   typedef std::vector<osg::ref_ptr<ColladaTechnique> > Technique_list;
   Technique_list _techs;

   // AMD added
   typedef std::map< std::string,  ColladaRTT * >  RTTNameMap;
   typedef std::map< unsigned int, ColladaRTT * >  RTTIndexMap;
   typedef std::vector< ColladaRTT * >  RTTVector;

   // AMD added
   RTTNameMap        m_RTTNameMap;
   RTTIndexMap       m_RTTIndexMap;
   RTTVector         m_RTTVector;

   // AMD added
   bool              m_PassIsRtt[128];

   // AMD added
   osg::Node         *m_pBaseGeometry;

   // AMD added
   float             m_fTimeCycle;
};

// INLINE METHODS

inline int ColladaEffectNode::getNumTechniques() const
{
   return static_cast<int>(_techs.size());
}

inline ColladaTechnique* ColladaEffectNode::getTechnique(int i)
{
   return _techs[i].get();
}

inline const ColladaTechnique* ColladaEffectNode::getTechnique(int i) const
{
   return _techs[i].get();
}

inline void ColladaEffectNode::addTechnique(ColladaTechnique* tech)
{
   _techs.push_back(tech);
}

#endif
