// AMD /////////////////////////////////////////////////////////////////
// DESCRIPTION: Collects all the things you need to create a render target
// into a single class.
// Copyright (c) 2008 Advanced Micro Devices, Inc. 
////////////////////////////////////////////////////////////////////////

#ifndef __COLLADA_RTT__
#define __COLLADA_RTT__

#include <osg/GLExtensions>
#include <osg/Node>
#include <osg/Geometry>
#include <osg/Notify>
#include <osg/MatrixTransform>
#include <osg/Texture2D>
#include <osg/TextureRectangle>
#include <osg/Stencil>
#include <osg/ColorMask>
#include <osg/Depth>
#include <osg/Billboard>
#include <osg/Material>
#include <osg/AnimationPath>
#include <osgGA/TrackballManipulator>
#include <osgGA/FlightManipulator>
#include <osgGA/DriveManipulator>
#include <osgUtil/SmoothingVisitor>
#include <osgDB/Registry>
#include <osgDB/ReadFile>
#include <osgDB/WriteFile>
#include <osgViewer/Viewer>
#include <iostream>

//=========================================================================================================
/// \ingroup 
/// \brief 
/// 
//=========================================================================================================
struct MyCameraPostDrawCallback : public osg::Camera::DrawCallback
{
   MyCameraPostDrawCallback( osg::Image* image, std::string strName )
      : _image ( image ),
       m_strName ( strName ) 
   {
   }

   virtual void operator () ( const osg::Camera& /*camera*/ ) const
   {
      std::string strFullName = m_strName + ".bmp";

      if ( osgDB::writeImageFile( *_image, strFullName ) )
      {
         printf ( "Writing image: %s\n", strFullName.c_str() );
      } 
      //_image->dirty();
   }

   osg::Image* _image;

   std::string m_strName ;
};


class ColladaRTT
{

public:

   ColladaRTT::ColladaRTT( std::string strName, unsigned int nIndex )
   {
      Setup();

      m_strTargetName = strName ;
      m_nPassIndex = nIndex ;
   }

   virtual ColladaRTT::~ColladaRTT()
   {    
   };

   osg::Group     * GetRootNode()
   {
      return m_pRoot;
   };

   osg::Texture   * GetTexture()
   {
      return m_pTexture;
   };

   osg::Camera   * GetCamera()
   {
      return m_pCamera;
   };

   void SetClearColor ( float fR, float fG, float fB, float fA );

   void SetClearDepth ( float fDepth );

   void AddTextureGeometry ( osg::Node * pTextureGeometry );

   bool CreateManualCamera(   osg::Vec3 & vLookAtPos,
                              osg::Vec3 & vEyePos,
                              osg::Vec3 & vEyeUp,
                              float       fZnear,
                              float       fZfar,
                              float       fXFOV,
                              float       fAspectRatio );

   osg::Texture * CreateTexture( unsigned nTexWidth, 
                                 unsigned nTexHeight,
                                 osg::Texture * pTexture = NULL );                                                                             

   std::string & GetName ()
   {
      return m_strTargetName;
   }  

   unsigned int GetPassIndex()
   {
      return m_nPassIndex ;
   }

   bool IsComplete();

   void CompleteNode();

   void SetMainViewLock( bool bVal );

   bool GetMainViewLock () 
   {
      return m_bMainViewLock; 
   }

   void HandleUpdate ( osg::NodeVisitor& nv, osg::Matrixf  & masterCameraVPW );

   static unsigned int m_nRTTCount;

   unsigned int   m_nMyRTTCount;

protected:

   void ColladaRTT::Setup();

private:

   osg::Node      * m_pTextureGeometry;
   osg::Group     * m_pRoot;
   osg::Camera    * m_pCamera;
   osg::Texture   * m_pTexture ;
   osg::Geode     * m_pGeode ;
   osg::Geometry  * m_pTestGeom ;

   // osg::Camera::FRAME_BUFFER_OBJECT
   osg::Camera::RenderTargetImplementation   m_RTI;

   bool           m_bUseImage;
   bool           m_bUseTextureRectangle;
   bool           m_bUseHDR;

   osg::Vec4      m_fClearColor ;
   float          m_fClearDepth ;

   // Name of the RTT surface
   std::string    m_strTargetName;

   bool           m_bCameraSet ;
   bool           m_bTextureSet ;
   bool           m_bClearColorSet ;
   bool           m_bClearDepthSet ;
   bool           m_bMainViewLock ;

   ///////////////////////////////////////////////////////////////
   // Cached values that are used to construct the final RTT node.
   osg::Vec3   m_vLookAtPos;
   osg::Vec3   m_vEyePos;
   osg::Vec3   m_vEyeUp;

   float       m_fZnear;
   float       m_fZfar;
   float       m_fXFOV;
   float       m_fAspectRatio;

   unsigned    m_nTexWidth;
   unsigned    m_nTexHeight; 

   unsigned int   m_nPassIndex;
};

#endif // __COLLADA_RTT__
