#include <osg/GLExtensions>
#include <osgUtil/CullVisitor>
#include <osgUtil/UpdateVisitor>
#include <osg/StateAttribute>
#include <osg/State>
#include <osg/Uniform>

#include "ColladaTechnique.h"
#include "ColladaEffectNode.h"

ColladaTechnique::ColladaTechnique()
:    osg::Referenced()

{
}

void ColladaTechnique::addPass( osg::StateSet * ss )
{
   if ( ss ) 
   {
      _passes.push_back( ss );
      ss->setRenderBinDetails( static_cast<int>( _passes.size() ), "RenderBin" );
   }
}

bool ColladaTechnique::validate(osg::State& state) const
{
   typedef std::vector<std::string> String_list;
   String_list extensions;

   getRequiredExtensions(extensions);

   for (String_list::const_iterator i=extensions.begin(); i!=extensions.end(); ++i) 
   {
      if (!osg::isGLExtensionSupported(state.getContextID(),i->c_str())) 
         return false;
   }

   return true;
}

void ColladaTechnique::traverse_implementation( osg::NodeVisitor& nv, ColladaEffectNode* fx )
{
   // define passes if necessary
   if ( _passes.empty() ) 
   {
      define_passes();
   }   

   // special actions must be taken if the node visitor is actually a CullVisitor
   osgUtil::CullVisitor *cv = dynamic_cast<osgUtil::CullVisitor *>(&nv);

   // traverse all passes
   for ( int i=0; i < getNumPasses(); ++i ) 
   {
      // AMD Added: Get the state set
      osg::StateSet *pSS = getPassStateSet ( i ) ;

      // push the i-th pass' StateSet if necessary
      if ( cv ) 
      {
         cv->pushStateSet( pSS );
      }

      // traverse the override node if defined, otherwise
      // traverse children as a Group would do
      osg::Node *override = getOverrideChild(i);
      if (override) 
      {
         override->accept(nv);
      } 
      else 
      {
         /****** AMD Added ************************************************************/
         bool bUpdateOnly = false;

         if ( nv.getVisitorType() == osg::NodeVisitor::UPDATE_VISITOR )
         {
            if ( pSS->requiresUpdateTraversal() )
            {
               // We still have to update the uniforms.
               pSS->runUpdateCallbacks( &nv );
            }
            bUpdateOnly = true ;
         }
         /****** End AMD Added ********************************************************/

         fx->inherited_traverse( nv, i );
      }

      // pop the StateSet if necessary
      if ( cv ) 
      {
         cv->popStateSet();
      }
   }        
}
