#include "daereader.h"

#include <dae.h>
#include <dae/daeSIDResolver.h>
#include <dae/domAny.h>
#include <dom/domCOLLADA.h>
#include <dom/domProfile_COMMON.h>

//HF
#include <dae/daeSmartRef.h>
#include <fstream>
#include <sstream>

//OSG includes
#include <osg/Texture1D>
#include <osg/Texture2D>
#include <osg/TextureRectangle>
#include <osg/TextureCubeMap>
#include <osg/Texture3D>
#include <osg/BlendFunc>
#include <osg/BlendEquation>
#include <osg/BlendColor>
#include <osg/AlphaFunc>
#include <osg/Depth>
#include <osg/ColorMask>
#include <osg/CullFace>
#include <osg/PolygonOffset>
#include <osg/PolygonMode> // AMD
#include <osg/LineWidth>   // AMD
#include <osgDB/Registry>
#include <osgDB/ReadFile>
#include <osgFX/Scribe>    // AMD
#include <osgFX/Cartoon>   // AMD
#include <osg/Uniform>     // AMD

#include "daeRMatrixUpdUtil.h"
#include "ColladaEffectNode.h" // AMD

// This is a experimental loader for the profile_GLSL of COLLADAFX
// Only the FIRST pass of the FIRST technique is loadeded

//Author: Heinrich Fink
//Date: 30.01.2007
//Mail: heinrich.fink@chello.at

// For example GLSL-files you might watch:
//		http://sourceforge.net/projects/collada-cg2glsl  Home of profile_CG to profile_GLSL conditioner
//		or
//		http://www.collada.org/owl/	COLLADA Test Model Bank

/**
 * Processes a profile_GLSL of a certain effect. 
 * NOTE: we HAVE to use a different stateset for each geometry, otherwise
 * WORLD-space uniform update-callbacks are not working correctly...
 */

using namespace osgdae;

// AMD added: Note this implemented used incorrectly and should not be used.
bool osgdae::daeReader::CheckEvaluateScene( domMaterial  *pDomMaterial, ColladaRTT *pRTT  )
{
   // Check for the visual_scene element
   if ( !m_pDoc->getScene()->getInstance_visual_scene() ) 
   {
      osg::notify( osg::WARN ) << "No visual scene found." << std::endl;
      return false;
   }

   bool bCameraSet = false;

   // Get the visual_scene element
   domInstanceWithExtra *ivs = m_pDoc->getScene()->getInstance_visual_scene();
   domVisual_scene *pVisualScene = daeSafeCast< domVisual_scene >( getElementFromURI( ivs->getUrl() ) );

   // For each evaluate scene
   int nVisualSceneCount = pVisualScene->getEvaluate_scene_array().getCount();
   for ( int i = 0 ; i < nVisualSceneCount ; i ++ )
   {
      // Get the evaluate scene.
      domVisual_scene::domEvaluate_scene *pEvalScene = pVisualScene->getEvaluate_scene_array().get(i);
      
      // For each render element
      int nRenderCount = pEvalScene->getRender_array().getCount();
      for ( int j = 0 ; j < nRenderCount; j++ )
      {
         // Get the render
         domVisual_scene::domEvaluate_scene::domRender *pRender = pEvalScene->getRender_array().get(j);
         if ( !pRender ) 
         {
            osg::notify( osg::WARN ) << "Evaluate scene render element not found in array." << std::endl;
            continue;
         }

         // Read the instance effect URL
         domInstance_effectRef rRenderInstanceEffect = pRender->getInstance_effect() ;
         xsAnyURI pRenderURI =  rRenderInstanceEffect->getUrl();
         
         // Compare URL to pDomMaterial's effect 
         domInstance_effectRef rMaterialInstanceEffect = pDomMaterial->getInstance_effect();
         xsAnyURI pMaterialURI = rMaterialInstanceEffect->getUrl();
               
         // If they are they same - load the camera node
         if ( pRenderURI == pMaterialURI )
         {
            // They are the same - go get the camera node
            xsAnyURI strCamName = pRender->getCamera_node();
            domNode *pCamNode = daeSafeCast< domNode >( getElementFromURI( pRender->getCamera_node() ) );
            if ( !pCamNode ) 
            {
               osg::notify( osg::WARN ) << "Could not locate a camera node for evaluate scene render element." << std::endl;
               continue;
            }

            // Get the lookat data for the camera.
            if ( 0 == pCamNode->getLookat_array().getCount() )
            {
               // If there is no Camera then we need to use the same values as the main camera.
               // Setup a camera with an update matrix.
               osg::notify( osg::WARN ) << "The camera node does not contain lookat data." << std::endl;
               continue;
            }

            // Load the first lookat data.
            domLookat *pLookat = pCamNode->getLookat_array().get ( 0 ) ;
            if ( !pLookat ) 
            {
               osg::notify( osg::WARN ) << "Could not load lookat data." << std::endl;
               continue;
            }

            // Get the matrix data
            domFloat3x3 fData = pLookat->getValue();
            osg::Vec3 vLookAtPos, vEyePos, vEyeUp;
            // The position of the eye is in the first 3 values
            vEyePos.set ( fData[0], fData[1],fData[2] );
            // The position looked at is in the second 3 values
            vLookAtPos.set ( fData[3], fData[4], fData[5] );
             // The up vector is in the third 3 values
            vEyeUp.set ( fData[6], fData[7], fData[8] );

            // Setup defaults vals.
            float fXFOV    = 45.0f;
            float fZNear   = 1.0f;
            float fZFar    = 1000.0f;
            float fAspectRatio = 1.0f;

            // Check to see if we have a camera instance.
            if ( 0 == pCamNode->getInstance_camera_array().getCount() )
            {
               osg::notify( osg::WARN ) << "No camera instance is specified in the camera node." << std::endl;
               continue;
            }
            // Now load the camera instance.
            domInstance_camera * pCameraInstance =  pCamNode->getInstance_camera_array().get(0);
            if ( !pCameraInstance ) 
            {
               osg::notify( osg::WARN ) << "Could not load camera instance" << std::endl;
               continue;
            }
            // Now go get the actual camera.
            domCamera *pCamera = daeSafeCast< domCamera >( getElementFromURI( pCameraInstance->getUrl() ) );
            if ( !pCamNode ) 
            {
               osg::notify( osg::WARN ) << "Could not locate a camera." << std::endl;
               continue;
            }
            // Get the optics
            domCamera::domOptics *pOptics = pCamera->getOptics();
            if ( !pOptics )
            {
               osg::notify( osg::WARN ) << "Could not load optics." << std::endl;
               continue;
            }
            // Get the common technique.
            domCamera::domOptics::domTechnique_common *pTechnique = pOptics->getTechnique_common();
            if ( !pTechnique )
            {
               osg::notify( osg::WARN ) << "Could not load technique." << std::endl;
               continue;
            }
            // Handle perspective.
            domCamera::domOptics::domTechnique_common::domPerspective *pPerspective = pTechnique->getPerspective();
            if ( !pPerspective )
            {
               osg::notify( osg::WARN ) << "Could not load perspective." << std::endl;
               continue;
            }
            else
            {
               fXFOV  = pPerspective->getXfov()->getValue();
               fZNear = pPerspective->getZnear()->getValue();
               fZFar  = pPerspective->getZfar()->getValue();
            }

            std::string strExtraValue;
            if ( daeReader::FindInExtra ( pCamera, "RenderMonkey", "MainViewLock", "", strExtraValue ) )
            {
               if ( strExtraValue == std::string ( "true" ) )
               {
                  pRTT->SetMainViewLock ( true ) ;
               }
            }

            // Create the RTT camera.
            pRTT->CreateManualCamera(  vLookAtPos, 
                                       vEyePos, 
                                       vEyeUp,
                                       fZNear,
                                       fZFar,
                                       fXFOV,
                                       fAspectRatio ) ;

            bCameraSet = true;
         }
      }
   }

   // If there is no camera define we may need to create one 
   //if ( !bCameraSet )
   //{
   //   pRTT->CreateManualCamera(  osg::Vec3 ( 0, 0, 0 ), 
   //                             osg::Vec3 ( 0, 180, 0 ), 
   //                              osg::Vec3 ( 0, 0, 1 ),
   //                              1.0f,
   //                              1000.0f,
   //                              45.0f,
   //                              1.0f ) ;
   //   pRTT->SetMainViewLock ( true ) ;
   //}

   return true;
}

// AMD many changes
ColladaEffectNode * osgdae::daeReader::processProfileGLSLUsingFX(    domProfile_GLSL            *pGlsl, 
                                                                     domInstance_materialRef    rDomMaterialRef, 
                                                                     domMaterial                *pDomMaterial,
                                                                     domEffect                  *pDomEffect,
                                                                     osg::Node                  *pParent )
{
	// TODO: also take the <technique_hint>-element of <instance_effect> into consideration
	// Get the first technique in effect
	domProfile_GLSL::domTechnique* technique = pGlsl->getTechnique_array()[0];

   // Create a new effect.
   ColladaEffectNode *pNewEffect = new ColladaEffectNode();
   if ( !pNewEffect )
   {
      osg::notify( osg::WARN ) << "Could not create ColladEffectNode" << std::endl;
      return NULL;
   }

   // Store for use in other methods.
   m_pCurrentColladaEffect = pNewEffect;
   // make sure thate the effect will receive update traversals
   pNewEffect->setUpdateCallback( new NodeDoesNothingCallback() );
   // Add the effect to the tree.
   pParent->asGroup()->addChild( pNewEffect );
   // Create a new Technique.
   ColladaTechnique *pNewTechnique = new  ColladaTechnique();
   // Add the technique to the effect.
   pNewEffect->addTechnique ( pNewTechnique );

   // Check for effect extras
   HandleEffectExtras ( pDomEffect, m_pCurrentColladaEffect );

   // For each pass
   unsigned int nPassCount = technique->getPass_array().getCount();
   for ( unsigned int nPassIndex = 0 ; nPassIndex < nPassCount ; nPassIndex++ )
   {
      osg::StateSet  *pStateSet  = NULL ;       
	   osg::Program   *sProgram   = NULL ;

      // Get the first pass
	   domProfile_GLSL::domTechnique::domPass* pDomPass = technique->getPass_array()[nPassIndex];

      GetStateSet( pDomPass, pStateSet, sProgram );

      {
		   //Load gl_pipeline_settings into current StateSet
		   for ( unsigned int i=0; i < pDomPass->getGl_pipeline_settings_array().getCount(); i++ ) 
         {
			   if ( !processGlPipelineSettings( pStateSet, pDomPass->getGl_pipeline_settings_array()[i]) ) 
            {
				   osg::notify( osg::WARN ) << "Could not process a Gl pipeline settings." << std::endl;
				   return NULL;
			   }
		   }

         // Check to see if this is an RTT pass (look for color targets).
         if( pDomPass->getColor_target_array().getCount() > 0 ) 
         {
            domFx_colortarget_common_Array & rColorTargetArray = pDomPass->getColor_target_array();
            for ( unsigned int nTargetIndex = 0 ; nTargetIndex < rColorTargetArray.getCount() ; nTargetIndex++ )
            {
               domFx_colortarget_commonRef rColorTarget = rColorTargetArray[nTargetIndex];

               // Locate the surface and extract the dimensions
               domGlsl_surface_type * pSurface = getSurfaceBySid ( rColorTarget->getValue() );
               if ( !pSurface )
               {
                  osg::notify( osg::WARN ) << "Could not locate surface for render target setup: " << rColorTarget->getValue() << std::endl;
                  return NULL;
               }

               ColladaRTT * pRTT = m_pCurrentColladaEffect->AddRenderTarget(  rColorTarget->getValue(), nPassIndex ) ; //rColorTarget->getIndex() );
               if ( !pRTT )
               {
                  osg::notify( osg::WARN ) << "Could not add render target: " << rColorTarget->getValue() << std::endl;
                  return NULL;
               }

               // Check to see if this effect is referenced in the evaluate_scene.
               // If it is it means that we need to read the camera data associated in the render camera_node.
               // The camera data will be used in the RTT object to set the view in the RTT pass.
               //CheckEvaluateScene( pDomMaterial, pRTT );

               // Setup the RTT camera. We assume that it will be the same as the main camera.
               pRTT->CreateManualCamera(  osg::Vec3 ( 0, 0, 0 ), 
                                          osg::Vec3 ( 0, 180, 0 ), 
                                          osg::Vec3 ( 0, 0, 1 ),
                                          1.0f,
                                          1000.0f,
                                          45.0f,
                                          1.0f ) ;
               pRTT->SetMainViewLock ( true ) ;

               // Get the size of the surface
               domFx_surface_common::domSize * pSize = pSurface->getSize();
               if ( !pSize )
               {
                  osg::notify( osg::WARN ) << "Could not locate size for render target setup: " << rColorTarget->getValue() << std::endl;
                  return NULL;
               }

               domInt3 nSize = pSize->getValue();
               osg::Texture *pNewTexture =  pRTT->CreateTexture( nSize[0], nSize[1] );   

               if ( !pNewTexture ) 
               {
                  osg::notify( osg::WARN ) << "Could not create render target texture for " << rColorTarget->getValue() << std::endl;
                  return NULL;
               }
               // Associate OSG textures (samplers) with surface name for later use
               m_RTTTextureMap[rColorTarget->getValue()] = pNewTexture;
            }
         }

         // Check for clear color that will be associated with the color targets
         if( pDomPass->getColor_clear_array().getCount() > 0 ) 
         {
            domFx_clearcolor_common_Array & rColorClearArray = pDomPass->getColor_clear_array();
            for ( unsigned int nTargetIndex = 0 ; nTargetIndex < rColorClearArray.getCount() ; nTargetIndex++ )
            {
               domFx_clearcolor_commonRef  rColorClear = rColorClearArray[nTargetIndex];
               ColladaRTT * pRTT = m_pCurrentColladaEffect->GetRenderTargetByIndex( nPassIndex ) ;
               if ( pRTT ) 
               {
                  domFx_color_common color = rColorClear->getValue();
                  float r = color.get(0);
                  float g = color.get(1);
                  float b = color.get(2);
                  float a = color.get(3);
                  pRTT->SetClearColor ( r, g, b, a );
               }
               else
               {
                  osg::notify( osg::WARN ) << "Could not locate RTT for clear_color at index " << rColorClear->getIndex() << std::endl;
               }
            }
         }

         // Check for clear color that will be associated with the color targets
         if( pDomPass->getDepth_clear_array().getCount() > 0 ) 
         {
            domFx_cleardepth_common_Array & rDepthClearArray = pDomPass->getDepth_clear_array();
            for ( unsigned int nTargetIndex = 0 ; nTargetIndex < rDepthClearArray.getCount() ; nTargetIndex++ )
            {
               domFx_cleardepth_commonRef  rDepthClear = rDepthClearArray[nTargetIndex];
               ColladaRTT * pRTT = m_pCurrentColladaEffect->GetRenderTargetByIndex( nPassIndex ) ;
               if ( pRTT ) 
               {
                  domFloat fDepth = rDepthClear->getValue();
                  pRTT->SetClearDepth ( fDepth );
               }
               else
               {
                  osg::notify( osg::WARN ) << "Could not locate RTT for clear_depth at index " << rDepthClear->getIndex() << std::endl;
               }
            }
         }
	   }

      // Process the shaders
	   domProfile_GLSL::domTechnique::domPass::domShader_Array shaderArray = pDomPass->getShader_array();
	   for ( unsigned int i = 0; i < shaderArray.getCount(); i++ ) 
      {
         std::string source = std::string("./") + std::string( shaderArray.get(i)->getName()->getSource() );
         daeElement* resolvedElement = ResolveElement ( source, shaderArray.get(i)->getParentElement() ) ;
		   if ( !resolvedElement ) 
         {
			   osg::notify( osg::WARN ) << "Could not resolve GLSL-Shader-Src." << std::endl;
			   return NULL;
		   }

         // Go get the shader code
         osg::Shader* pShader = LoadShader ( resolvedElement, shaderArray.get(i)->getStage() );
		   if ( pShader == NULL )
         {
            return NULL;
         }
			
		   //Add shader to program
		   if ( !sProgram->addShader( pShader ) ) 
         {
			   osg::notify( osg::WARN ) << "Could not add shader to program." << std::endl;
			   return NULL;
		   }

         // Add the vertex attributes.
         AddVertexAttributes ( sProgram, rDomMaterialRef->getBind_vertex_input_array() );

		   // Generate the uniforms based on the bind-array of current domShader
		   if ( !processBindArray(    pStateSet, 
                                    shaderArray.get(i), 
                                    pDomPass, 
                                    pDomMaterial,
                                    m_pCurrentColladaEffect ) ) 
         {
			   osg::notify( osg::WARN ) << "Could not process bind-array of profile_GLSL" << std::endl;
			   return NULL;
		   }	
      }// End Shaders
   
      {
	      //Add program to state
	      pStateSet->setAttributeAndModes( sProgram, osg::StateAttribute::ON );
	      //Add to effect-cache
         m_passGLSLMap[pDomPass] = pStateSet;
      }

      // Add the pass to the technique
      pNewTechnique->cachePass( pStateSet );

      pStateSet->setUpdateCallback( new StateSetDoesNothingCallback() );
   }// End Pass

	return pNewEffect;
}
// End AMD change


bool osgdae::daeReader::processGlPipelineSettings( osg::StateSet* ss, domGl_pipeline_settings* pipelineSetting ) {

	//TODO: also load if referenced via "param"-attribute

   // GL_AlphaFunction, NOTE: only the "value" is properly handled yet
   if ( pipelineSetting->getAlpha_func() != NULL ) 
   {
		osg::AlphaFunc* alphaFunc = new osg::AlphaFunc();

		if ( pipelineSetting->getAlpha_func()->getValue()->getValue()!= NULL ) 
      {
			alphaFunc->setReferenceValue( pipelineSetting->getAlpha_func()->getValue()->getValue() );
		} 
      else 
      {
			osg::notify( osg::WARN ) << "No Value for AlphaFunc -> Value, param attribute not supported yet " << std::endl;
			return false;
		}

		if ( pipelineSetting->getAlpha_func()->getFunc()->getValue() != NULL ) 
      {
			domGl_func_type funcType = pipelineSetting->getAlpha_func()->getFunc()->getValue();

			if (funcType == GL_FUNC_TYPE_ALWAYS) 
				alphaFunc->setFunction( osg::AlphaFunc::ALWAYS );
			else if (funcType == GL_FUNC_TYPE_EQUAL)
				alphaFunc->setFunction( osg::AlphaFunc::EQUAL );
			else if (funcType == GL_FUNC_TYPE_GEQUAL )
				alphaFunc->setFunction( osg::AlphaFunc::GEQUAL );
			else if (funcType == GL_FUNC_TYPE_GREATER )
				alphaFunc->setFunction( osg::AlphaFunc::GREATER );
			else if (funcType == GL_FUNC_TYPE_LEQUAL )
				alphaFunc->setFunction( osg::AlphaFunc::LEQUAL );
			else if (funcType == GL_FUNC_TYPE_LESS )
				alphaFunc->setFunction( osg::AlphaFunc::LESS );
			else if (funcType == GL_FUNC_TYPE_NEVER )
				alphaFunc->setFunction( osg::AlphaFunc::NEVER );
			else if (funcType == GL_FUNC_TYPE_NOTEQUAL )
				alphaFunc->setFunction( osg::AlphaFunc::NOTEQUAL );
			else 
         {
				osg::notify( osg::WARN ) << "Undefined AlphaFunct function value" << std::endl;
				return false;
			}
		} 
      else 
      {
			osg::notify( osg::WARN ) << "No Value for AlphaFunc -> Func, param attribute not supported yet " << std::endl;
			return false;
		}
		//Add to stateset and activate its mode
		ss->setAttribute( alphaFunc, osg::StateAttribute::ON );
		//NOTE: This has also to be enabled via <alpha_test_enable>
	}
   // GL_BlendEquation
	else if ( pipelineSetting->getBlend_func() != NULL ) 
   {
		osg::BlendFunc* blendFunc = new osg::BlendFunc();
		
		if (pipelineSetting->getBlend_func()->getSrc()->getValue() != NULL) 
      {
			domGl_blend_type blendType = pipelineSetting->getBlend_func()->getSrc()->getValue();
			blendFunc->setSource( m_blendFuncModeMap[blendType] );
		} 
      else 
      {
			osg::notify( osg::WARN ) << "No Value for <blend_func> -> SRC, param attribute not supported yet." << std::endl;
			return false;
		}

		if ( pipelineSetting->getBlend_func()->getDest()->getValue() != NULL ) 
      {
			domGl_blend_type blendType = pipelineSetting->getBlend_func()->getDest()->getValue();
			blendFunc->setDestination( m_blendFuncModeMap[blendType] );
		} 
      else 
      {
			osg::notify( osg::WARN ) << "No Value for <blend_func> -> DST, param attribute not supported yet." << std::endl;
			return false;
		}

		ss->setAttribute( blendFunc, osg::StateAttribute::ON );
	}
   // AMD change 
   // <blend_func_separate>
	else if ( pipelineSetting->getBlend_func_separate() != NULL ) 
   {
		osg::BlendFunc* blendFuncSep = new osg::BlendFunc();
		
      domGl_blend_type  destAlpha   = GL_BLEND_TYPE_ZERO, 
                        srcAlpha    = GL_BLEND_TYPE_ONE,
                        destRgb     = GL_BLEND_TYPE_ZERO, 
                        srCRgb      = GL_BLEND_TYPE_ONE;

      if ( pipelineSetting->getBlend_func_separate()->getDest_alpha() )
      {
         destAlpha = pipelineSetting->getBlend_func_separate()->getDest_alpha()->getValue();
      }

      if ( pipelineSetting->getBlend_func_separate()->getSrc_alpha() )
      {
		   srcAlpha = pipelineSetting->getBlend_func_separate()->getSrc_alpha()->getValue();
      }

      if ( pipelineSetting->getBlend_func_separate()->getDest_rgb() )
      {
		   destRgb = pipelineSetting->getBlend_func_separate()->getDest_rgb()->getValue();
      }

      // AMD change
      if ( pipelineSetting->getBlend_func_separate()->getSrc_rgb() )
      {
         srCRgb = pipelineSetting->getBlend_func_separate()->getSrc_rgb()->getValue();
      }

		blendFuncSep->setDestinationAlpha( m_blendFuncModeMap[destAlpha] );
		blendFuncSep->setSourceAlpha( m_blendFuncModeMap[srcAlpha] );
		blendFuncSep->setDestinationRGB( m_blendFuncModeMap[destRgb] );
		blendFuncSep->setSourceRGB( m_blendFuncModeMap[srCRgb] );

		ss->setAttribute( blendFuncSep, osg::StateAttribute::ON );
	}
   // GL_BlendEquation
	else if ( pipelineSetting->getBlend_equation() != NULL ) 
   {
		osg::BlendEquation* blendEquation = new osg::BlendEquation();

		if ( pipelineSetting->getBlend_equation()->getValue() != NULL ) 
      {
			domGl_blend_equation_type blendEqType = pipelineSetting->getBlend_equation()->getValue();

			if (blendEqType == GL_BLEND_EQUATION_TYPE_FUNC_ADD)
				blendEquation->setEquation( osg::BlendEquation::FUNC_ADD );
			else if (blendEqType == GL_BLEND_EQUATION_TYPE_FUNC_REVERSE_SUBTRACT)
				blendEquation->setEquation( osg::BlendEquation::FUNC_REVERSE_SUBTRACT );
			else if (blendEqType == GL_BLEND_EQUATION_TYPE_FUNC_SUBTRACT)
				blendEquation->setEquation( osg::BlendEquation::FUNC_SUBTRACT );
			else if (blendEqType == GL_BLEND_EQUATION_TYPE_MAX)
				blendEquation->setEquation( osg::BlendEquation::RGBA_MAX );
			else if (blendEqType == GL_BLEND_EQUATION_TYPE_MIN)
				blendEquation->setEquation( osg::BlendEquation::RGBA_MIN );
		} 

		ss->setAttribute( blendEquation, osg::StateAttribute::ON );
	}
   // GL_BlendEquation
	else if ( pipelineSetting->getBlend_equation_separate() != NULL ) 
   {
		osg::BlendEquation* blendEquation = new osg::BlendEquation();

		//Check if all elements have a 'value' attribtue
		if (  pipelineSetting->getBlend_equation_separate()->getRgb()->getValue() == NULL || 
			   pipelineSetting->getBlend_equation_separate()->getAlpha()->getValue() == NULL ) 
      {
				osg::notify( osg::WARN ) << "No all elements of <blend_equation_separate> have a 'value' attribute, 'param' is not supported yet." << std::endl;
				return false;
		}

		domGl_blend_equation_type rgbBlendEqType   = pipelineSetting->getBlend_equation_separate()->getRgb()->getValue();
		domGl_blend_equation_type alphaBlendEqType = pipelineSetting->getBlend_equation_separate()->getAlpha()->getValue();

		//NOTE: Seems like separation works only when using MIN or MAX
		//Setting RGB BlendType
		if (rgbBlendEqType == GL_BLEND_EQUATION_TYPE_FUNC_ADD)
			blendEquation->setEquation( osg::BlendEquation::FUNC_ADD );
		else if (rgbBlendEqType == GL_BLEND_EQUATION_TYPE_FUNC_REVERSE_SUBTRACT)
			blendEquation->setEquation( osg::BlendEquation::FUNC_REVERSE_SUBTRACT );
		else if (rgbBlendEqType == GL_BLEND_EQUATION_TYPE_FUNC_SUBTRACT)
			blendEquation->setEquation( osg::BlendEquation::FUNC_SUBTRACT );
		else if (rgbBlendEqType == GL_BLEND_EQUATION_TYPE_MAX)
			blendEquation->setEquation( osg::BlendEquation::RGBA_MAX );
		else if (rgbBlendEqType == GL_BLEND_EQUATION_TYPE_MIN)
			blendEquation->setEquation( osg::BlendEquation::RGBA_MIN );

		//We check only for MIN MAX in Alpha
		if (alphaBlendEqType == GL_BLEND_EQUATION_TYPE_MIN)
			blendEquation->setEquation( osg::BlendEquation::ALPHA_MIN );
		else if (alphaBlendEqType == GL_BLEND_EQUATION_TYPE_MAX)
			blendEquation->setEquation( osg::BlendEquation::ALPHA_MAX );

		ss->setAttribute( blendEquation, osg::StateAttribute::ON );
	}
        // AMD change
	// GL_CullMode <cull_face_enable>
	else if ( pipelineSetting->getCull_face_enable() != NULL) 
   {
		if ( pipelineSetting->getCull_face_enable()->getValue() )
			ss->setMode( GL_CULL_FACE, GL_TRUE );
		else 
			ss->setMode( GL_CULL_FACE, GL_FALSE );
	}
   // GL_CullMode <cull_face>
	else if ( pipelineSetting->getCull_face() != NULL ) 
   {
		osg::CullFace* cullFace = new osg::CullFace();

		if ( pipelineSetting->getCull_face()->getValue() != NULL ) 
      {
			domGl_face_type faceType = pipelineSetting->getCull_face()->getValue();
			if ( faceType == GL_FACE_TYPE_BACK )
				cullFace->setMode( osg::CullFace::BACK );
			else if ( faceType == GL_FACE_TYPE_FRONT )
				cullFace->setMode( osg::CullFace::FRONT );
			else if ( faceType == GL_FACE_TYPE_FRONT_AND_BACK )
				cullFace->setMode( osg::CullFace::FRONT_AND_BACK );
		} 
      else 
      {
			osg::notify( osg::WARN ) << "gl_pipeline_setting <cull_face> has no 'value' attribute, param references are not supported yet." << std::endl;
			return false;
		}

		ss->setAttribute( cullFace, osg::StateAttribute::ON);
	}
   // GL_DepthFunction
	else if ( pipelineSetting->getDepth_func() != NULL ) 
   {
		osg::Depth* osgDepth = NULL;
		osg::StateAttribute* sa = ss->getAttribute(osg::StateAttribute::DEPTH);
		 
		if (sa == NULL)
			osgDepth = new osg::Depth();
		else 
			osgDepth = static_cast<osg::Depth* >( sa );

		if (pipelineSetting->getDepth_func()->getValue() != NULL) 
      {
			domGl_func_type funcType = pipelineSetting->getDepth_func()->getValue();

			if ( funcType == GL_FUNC_TYPE_ALWAYS )
				osgDepth->setFunction( osg::Depth::ALWAYS );
			else if ( funcType == GL_FUNC_TYPE_EQUAL )
				osgDepth->setFunction( osg::Depth::EQUAL );
			else if ( funcType == GL_FUNC_TYPE_GEQUAL )
				osgDepth->setFunction( osg::Depth::GEQUAL );
			else if ( funcType == GL_FUNC_TYPE_GREATER )
				osgDepth->setFunction( osg::Depth::GREATER );
			else if ( funcType == GL_FUNC_TYPE_LEQUAL )
				osgDepth->setFunction( osg::Depth::LEQUAL );
			else if ( funcType == GL_FUNC_TYPE_LESS )
				osgDepth->setFunction( osg::Depth::LESS );
			else if ( funcType == GL_FUNC_TYPE_NEVER )
				osgDepth->setFunction( osg::Depth::NEVER );
			else if ( funcType == GL_FUNC_TYPE_NOTEQUAL)
				osgDepth->setFunction( osg::Depth::NOTEQUAL );
		} 
      else 
      {
			osg::notify( osg::WARN ) << "gl_pipeline_setting <depth_func> has no 'value' attribute, param references are not supported yet." << std::endl;
			return false;
		}

		//NOTE: This should work also if osgDepth is already set as an attribute of this stateset
		ss->setAttribute( osgDepth, osg::StateAttribute::ON );
	}
	// GL_BlendColor <blend_color>
	else if ( pipelineSetting->getBlend_color() != NULL ) 
   {
		osg::BlendColor* blendColor = new osg::BlendColor();

		blendColor->setConstantColor( 
				osg::Vec4(
					pipelineSetting->getBlend_color()->getValue()[0],
					pipelineSetting->getBlend_color()->getValue()[1],
					pipelineSetting->getBlend_color()->getValue()[2],
					pipelineSetting->getBlend_color()->getValue()[3]
					) 
				);

		ss->setAttribute( blendColor, osg::StateAttribute::ON );
	}
	// TODO: I guess this should go to osg::Camera, or by creating a new osg::ClearNode
	// However, setting the clear-depth might be very important...
	// <depth_range>
	else if ( pipelineSetting->getDepth_range() != NULL ) 
   {
		osg::StateAttribute* sa = ss->getAttribute(osg::StateAttribute::DEPTH);
		osg::Depth* osgDepth = NULL;

		if ( sa == NULL )
      {
			osgDepth = new osg::Depth();
      }
		else
      {
			osgDepth = static_cast<osg::Depth* >( sa );
      }

		osgDepth->setRange(  static_cast<double>( pipelineSetting->getDepth_range()->getValue()[0]),
			                  static_cast<double>( pipelineSetting->getDepth_range()->getValue()[1]) );

		ss->setAttribute( osgDepth, osg::StateAttribute::ON );
	}
	// GL_WriteMaskColor <color_mask>
	else if ( pipelineSetting->getColor_mask() != NULL ) 
   {
		osg::ColorMask* osgColorMask = new osg::ColorMask();

		osgColorMask->setMask(
			pipelineSetting->getColor_mask()->getValue().get(0),
			pipelineSetting->getColor_mask()->getValue().get(1),
			pipelineSetting->getColor_mask()->getValue().get(2),
			pipelineSetting->getColor_mask()->getValue().get(3)
			);	

		ss->setAttribute( osgColorMask, osg::StateAttribute::ON );
	}
	// GL_WriteMaskDepth <depth_mask>
	else if ( pipelineSetting->getDepth_mask() != NULL ) 
   {
		osg::StateAttribute* sa = ss->getAttribute(osg::StateAttribute::DEPTH);
		osg::Depth* osgDepth = NULL;

		if (sa == NULL)
			osgDepth = new osg::Depth();
		else 
			osgDepth = static_cast<osg::Depth* >( sa );

		osgDepth->setWriteMask( pipelineSetting->getDepth_mask()->getValue() );
		ss->setAttribute( osgDepth, osg::StateAttribute::ON );
	}
	// GL_PolyOffsetFactor, GL_PolyOffsetunits <polgyon_offset>
	else if ( pipelineSetting->getPolygon_offset() != NULL ) 
   {
		osg::PolygonOffset* osgPolyOffset = NULL;
      domGl_pipeline_settings::domPolygon_offset *pOffset = pipelineSetting->getPolygon_offset();
      domFloat2 fVals = pOffset->getValue();

      if ( fVals.getCount() > 1 )
      {
         osgPolyOffset = new osg::PolygonOffset( fVals[0], fVals[1] );
      }
      else
      {
         osgPolyOffset = new osg::PolygonOffset( 0.0f, 0.0f );
      }
		
		ss->setAttribute( osgPolyOffset, osg::StateAttribute::ON );
	}
	// GL_AlphaEnable <alpha_test_enable>
	else if ( pipelineSetting->getAlpha_test_enable() != NULL ) 
   {
		if ( pipelineSetting->getAlpha_test_enable()->getValue() != NULL )
      {
			ss->setMode( GL_ALPHA_TEST, pipelineSetting->getAlpha_test_enable()->getValue() );
      }
	}
	// GL_BlendEnable <blend_enable>
	else if ( pipelineSetting->getBlend_enable() != NULL ) 
   {
		if ( pipelineSetting->getBlend_enable()->getValue() ) 
      {
			ss->setMode( GL_BLEND, GL_TRUE );
			//Sort for correct transparency rendering
			ss->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );
			ss->setRenderBinDetails( 10, "DepthSortedBin" );
		} 
      else
      {
			ss->setMode( GL_BLEND, GL_FALSE );
      }
	}
	// GL_DepthEnable <depth_test_enable>
	else if ( pipelineSetting->getDepth_test_enable() != NULL ) 
   {
		if ( pipelineSetting->getDepth_test_enable()->getValue() )
			ss->setMode( GL_DEPTH_TEST, GL_TRUE );
		else 
			ss->setMode( GL_DEPTH_TEST, GL_FALSE );
	}
        // AMD change
	// GL_LineSmooth <line_smooth_enable>
	else if ( pipelineSetting->getLine_smooth_enable() != NULL ) 
   {
		if ( pipelineSetting->getLine_smooth_enable()->getValue() )
			ss->setMode( GL_LINE_SMOOTH, GL_TRUE );
		else 
			ss->setMode( GL_LINE_SMOOTH, GL_FALSE );
	}
	// GL_PolySmooth <polygon_smooth_enable>
	else if ( pipelineSetting->getPolygon_smooth_enable() != NULL ) 
   {
		if ( pipelineSetting->getPolygon_smooth_enable()->getValue() )
			ss->setMode( GL_POLYGON_SMOOTH, GL_TRUE );
		else 
			ss->setMode( GL_POLYGON_SMOOTH, GL_FALSE );
	}
	// GL_PolyOffsetFill <polygon_offset_fill_enable>
	else if ( pipelineSetting->getPolygon_offset_fill_enable() != NULL ) 
   {
		if ( pipelineSetting->getPolygon_offset_fill_enable()->getValue() )
			ss->setMode( GL_POLYGON_OFFSET_FILL, GL_TRUE );
		else 
			ss->setMode( GL_POLYGON_OFFSET_FILL, GL_FALSE );
	}
	// GL_PolyOffsetLine <polygon_offset_line_enable>
	else if ( pipelineSetting->getPolygon_offset_line_enable() != NULL ) 
   {
		if ( pipelineSetting->getPolygon_offset_line_enable()->getValue() )
			ss->setMode( GL_POLYGON_OFFSET_LINE, GL_TRUE );
		else 
			ss->setMode( GL_POLYGON_OFFSET_LINE, GL_FALSE );
	}
	// GL_PolyOffsetPoint <polygon_offset_point enable>
	else if ( pipelineSetting->getPolygon_offset_point_enable() != NULL ) 
   {
		if ( pipelineSetting->getPolygon_offset_point_enable()->getValue() )
			ss->setMode( GL_POLYGON_OFFSET_POINT, GL_TRUE );
		else 
			ss->setMode( GL_POLYGON_OFFSET_POINT, GL_FALSE );
	}
   // AMD change
   // GL_PolyFrontMode, GL_PolyBackMode <polygon_mode>
   else if ( pipelineSetting->getPolygon_mode() != NULL )
   {
      // Check the face mode exists 
      if ( pipelineSetting->getPolygon_mode()->getFace() == NULL  )
      {
         osg::notify( osg::WARN ) << "gl_pipeline_setting <polygon_mode> has no face mode attribute, param references are not supported yet." << std::endl;
			return false;
		}

      // Check the face mode value exists
      if ( pipelineSetting->getPolygon_mode()->getFace()->getValue() == NULL  )
      {
         osg::notify( osg::WARN ) << "gl_pipeline_setting <polygon_mode> face has no 'value' attribute, param references are not supported yet." << std::endl;
			return false;
		}

      // Get the face type. (front, back, or both)
      domGl_face_type eFaceType = pipelineSetting->getPolygon_mode()->getFace()->getValue();
      osg::PolygonMode::Face osgFace;

      switch ( eFaceType )
      {
      case GL_FACE_TYPE_FRONT :
         osgFace = osg::PolygonMode::FRONT;
         break;

      case GL_FACE_TYPE_BACK : 
         osgFace = osg::PolygonMode::BACK;
         break;

      case GL_FACE_TYPE_FRONT_AND_BACK : 
         osgFace = osg::PolygonMode::FRONT_AND_BACK;
         break;

      default:
         osg::notify( osg::WARN ) << "gl_pipeline_setting <polygon_mode> Face has unknown value." << std::endl;
		   return false;
      }
      
      // Check to see if we have a polygon mode
      if ( pipelineSetting->getPolygon_mode()->getMode() == NULL  )
      {
         osg::notify( osg::WARN ) << "gl_pipeline_setting <polygon_mode> has no mode attribute, param references are not supported yet." << std::endl;
			return false;
      }

      // Check to see if we have a polygon mode type ( lines, points, fill).
      if ( pipelineSetting->getPolygon_mode()->getMode()->getValue() == NULL  )
      {
         osg::notify( osg::WARN ) << "gl_pipeline_setting <polygon_mode> mode has no 'value' attribute, param references are not supported yet." << std::endl;
			return false;
      }

      osg::PolygonMode* polygonMode = new osg::PolygonMode();
	   domGl_polygon_mode_type eModeType = pipelineSetting->getPolygon_mode()->getMode()->getValue();
      osg::PolygonMode::Mode osgMode;

      switch ( eModeType )
      {
      case GL_POLYGON_MODE_TYPE_POINT:
         osgMode = osg::PolygonMode::POINT;
         break;

      case GL_POLYGON_MODE_TYPE_LINE:
         osgMode = osg::PolygonMode::LINE;
         break;

      case GL_POLYGON_MODE_TYPE_FILL:
         osgMode = osg::PolygonMode::FILL;
         break;

      default:
         osg::notify( osg::WARN ) << "gl_pipeline_setting <polygon_mode> Mode has unknown value." << std::endl;
	      return false;
      }

      polygonMode->setMode( osgFace, osgMode );
      ss->setAttribute( polygonMode, osg::StateAttribute::ON);
   }
   // GL_LineWidth <line_width>
   else if ( pipelineSetting->getLine_width() != NULL )
   {
      if ( pipelineSetting->getLine_width()->getValue() == NULL )
      {
         osg::notify( osg::WARN ) << "gl_pipeline_setting <line_width> has no 'value' attribute, param references are not supported yet." << std::endl;
			return false;
      }

      domFloat fWidth = pipelineSetting->getLine_width()->getValue();
      osg::LineWidth* lineWidth = new osg::LineWidth();
      lineWidth->setWidth ( fWidth ) ;
      ss->setAttribute( lineWidth, osg::StateAttribute::ON);
   }

	//TODO: Finish ALL gl_pipeline_settings (easy, but lots of typing...)
   else if ( pipelineSetting->getDepth_bounds() != NULL ) 
   {
      // Handle Depth_bounds here.
      osg::notify( osg::WARN ) << "Please implement depth bounds" << std::endl;
   }
   else if ( pipelineSetting->getFront_face() != NULL ) 
   {
      // Handle front face here.
      osg::notify( osg::WARN ) << "Please implement front face" << std::endl;
   }
   else if ( pipelineSetting->getClear_depth() != NULL ) 
   {
      // Handle clear depth here.
      osg::notify( osg::WARN ) << "Please implement clear depth" << std::endl;
   }
   else if ( pipelineSetting->getShade_model() != NULL ) 
   {
      // Handle shade model here.
      osg::notify( osg::WARN ) << "Please implement shade model" << std::endl;
   }
 
	return true;
}
// End AMD change


// AMD add
bool osgdae::daeReader::OveriddenByMaterial (   //domProfile_GLSL::domTechnique::domPass::domShader::domBind* pBindEl,
                                                const std::string &  rStrParamName, 
                                                domMaterial       *  pDomMaterial,
                                                unsigned int      &  nParamIndex )
{
   assert ( pDomMaterial ) ;
   //assert ( pBindEl );

   domInstance_effectRef rInstanceEffect = pDomMaterial->getInstance_effect();

   if ( rInstanceEffect == NULL ) 
   {
		osg::notify( osg::WARN ) << "Error getting <technique> from <bind>-element. " << std::endl;
		return false;
	}
   
   //std::string strSetParamName = pBindEl->getParam()->getRef();
   
   for ( unsigned int i = 0 ; i < rInstanceEffect->getSetparam_array().getCount(); i++ )
   {
      std::string strSetParamName =  rInstanceEffect->getSetparam_array()[i]->getRef();
		if ( rStrParamName == strSetParamName ) 
      {
			// TODO: also check for semantics in this setparam - array
         nParamIndex = i;
         return true ;
      }
   }

   return false;
}
// End AMD add

// AMD changes
//Resolves params in <bind>-array, calls processGLSLParamType, and
//attaches the resulting osg::Uniforms to the current stateset
bool osgdae::daeReader::processBindArray( osg::StateSet* ss, 
                                          domProfile_GLSL::domTechnique::domPass::domShader * shaderEl,
                                          domProfile_GLSL::domTechnique::domPass * pDomPass,
                                          domMaterial       *  pDomMaterial,
                                          ColladaEffectNode *  pEffect ) 
{	
	//NOTE: if updateSemanticBindsOnly is set, we just create 
	//new uniforms for the callbacks for some float4x4's
   assert ( pEffect ) ;

   // Reset the member pointers that will be updated on each call to this method.
   m_pCurrentDomMaterial  = pDomMaterial ;

	ParamToUniformMap loadedUniforms;
	int texUnitCounter = 0;
	//Go through each element in <bind>-array, resolve the param-element, 
	//create osg::uniform for that
   unsigned int nBindCount = shaderEl->getBind_array().getCount() ;
	for ( unsigned int i = 0; i < nBindCount; i++ ) 
   {
		domProfile_GLSL::domTechnique::domPass::domShader::domBind* pBindEl = shaderEl->getBind_array().get(i);

		// We either have a <param> element with a reference, 
		// or a constant glsl_type
		osg::Uniform* glslUniform = NULL;

      // Process the common Params
		if ( pBindEl->getParam() != NULL ) 
      {
			//Resolve to newparam in steps
			//Resolve in steps to technique->profile->effect scope
			std::string source = std::string("./") + std::string( pBindEl->getParam()->getRef() );
         daeElement* resolvedElement = ResolveElement ( source, shaderEl->getParentElement() );
         if ( resolvedElement == NULL ) 
         {
				osg::notify( osg::WARN ) << "Could not resolve <newparam> from <bind>-array." << source <<std::endl;
				return false;
			}

         // Figure out if we have a glsl or common param
			domGlsl_newparam        * glslNewParam    = daeSafeCast<domGlsl_newparam >( resolvedElement );
			domFx_newparam_common   * commonNewParam  = daeSafeCast<domFx_newparam_common >( resolvedElement );

         std::string sid;

         // Handle overrides at the technique level.
         unsigned int nTechniqueParamIndex = 0;
         unsigned int nMaterialParamIndex  = 0;

         // Check to see if the bound uniform is overriden by data in the material description (setparams)
         bool bOveriddenByMaterial  = OveriddenByMaterial ( pBindEl->getParam()->getRef(), m_pCurrentDomMaterial,  nMaterialParamIndex ) ;
         if ( bOveriddenByMaterial )
         {
            // Check that the current material has an instance effect (it should). 
            domInstance_effectRef rInstanceEffect = pDomMaterial->getInstance_effect();
            if ( rInstanceEffect == NULL ) 
            {
               osg::notify( osg::WARN ) << "Error getting <instance_effect> from <material>-element. " << std::endl;
               return false;
            }
            
            // Process the common data. We override the newparam with the setparam data.
            if ( !processFxCommonParamType( 
                        rInstanceEffect->getSetparam_array()[nMaterialParamIndex]->getFx_basic_type_common(), // Data to use 
                        pBindEl->getSymbol(), // Name of the uniform in the shader code.
                        glslUniform ) ) // The output uniform.
            {
                osg::notify( osg::WARN ) << "Creating osg::UNiform's out of <setparam> in <material> failed" << std::endl;
                  return false;
            }

            // Assign the id of the newparam
            sid = pBindEl->getParam()->getRef();
         }
         else // The uniform data (newparam) is not overridden with a setparam in the material
         {	
            // Check if we have a GLSL param
				if ( glslNewParam != NULL ) 
            {
					// Process the GLSL data
					if ( !processGlslParamType( glslNewParam->getGlsl_param_type(), pBindEl->getSymbol(), glslUniform )) 
               {
						osg::notify( osg::WARN ) << "Process GLSL param type failed." << std::endl;
						return false;
					}
					sid = std::string( glslNewParam->getSid() );
				} 
            // Or if we have a common param.
            else if ( commonNewParam != NULL ) 
            {
					if ( !processFxCommonParamType( commonNewParam->getFx_basic_type_common(), pBindEl->getSymbol(), glslUniform ) ) 
               {
						osg::notify( osg::WARN ) << "Process FX COMMON param type failed^" << std::endl;
						return false;
					}
					sid = std::string( commonNewParam->getSid() );
				} 
            else 
            {
					osg::notify( osg::WARN ) << "Incompatible resolved type. Critical Error. " << std::endl;
					return false;
				}
         }

         // View Matrices use setUserData (per frame only)
         // Transforms (world matrices) use callbacks (per object)
         // Non matrix Unifoms use callbacks

			// Process the semantics
			if ( glslNewParam && glslNewParam->getSemantic() !=  NULL ) 
         {	
				//To upper case for case insensitiv comparison
				std::string semanticString( glslNewParam->getSemantic()->getValue() );
				for ( unsigned int j=0; j<semanticString.size(); j++ ) 
            {
					if ( isalpha( semanticString.at(j) ))
               {
						semanticString[j] = toupper( semanticString[j] );
               }
				}			

            // This code no longer required as as we only create/modify uniforms one time.

				//ss->removeUniform(ss->getUniform( std::string( pBindEl->getSymbol() ) ));
				//if ( !processGlslParamType( glslNewParam->getGlsl_param_type(), pprocessbindEl->getSymbol(), glslUniform )) 
            //{
				//	osg::notify( osg::WARN ) << "Updating a float uniform semantic failed" << std::endl;
				//	return false;
				//}

            if ( semanticString == "TIME0_X") 
            {       
					glslUniform->setUpdateCallback( new Time_ZERO_X_Callback() );
				} 
				else if ( semanticString == "TIME0_1") 
            {       
               glslUniform->setUpdateCallback( new Time_ZERO_ONE_Callback( pEffect->GetTimeCycle() ) );
				} 
				else if ( semanticString == "TIME0_2PI") 
            {       
					glslUniform->setUpdateCallback( new Time_ZERO_2PI_Callback( pEffect->GetTimeCycle() ) );
				} 
				else if ( semanticString == "COSTIME0_2PI") 
            {       
					glslUniform->setUpdateCallback( new Time_COS_ZERO_2PI_Callback( pEffect->GetTimeCycle() ) );
				} 
				else if ( semanticString == "SINTIME0_2PI" ) 
            {       
					glslUniform->setUpdateCallback( new Time_SIN_ZERO_2PI_Callback( pEffect->GetTimeCycle() ) );
				} 
				else if ( semanticString == "SINTIME0_X") 
            {       
					glslUniform->setUpdateCallback( new Time_SIN_ZERO_X_Callback() );
				} 
				else if ( semanticString == "COSTIME0_X") 
            {       
					glslUniform->setUpdateCallback( new Time_COS_ZERO_X_Callback() );
				} 
				else if ( semanticString == "VIEWPOSITION") 
            {                  
               // This method does update but the value seems to be constant
					//ViewPositionExtractor* viewPositionExtractor = new ViewPositionExtractor( glslUniform );
					//ss->setAttribute( viewPositionExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );
					//glslUniform->setUserData( viewPositionExtractor );
				} 
				else if ( semanticString == "PASSINDEX" ) 
            {  
               // Nothing is required for PassIndex as we cannot update a uniform
               // between passes anyway.
				} 
				else if ( semanticString == "VIEWDIRECTION" ) 
            {                  
               // This method does update but the value seems to be constant
					//ViewDirectionExtractor* viewDirectionExtractor = new ViewDirectionExtractor( glslUniform );
					//ss->setAttribute( viewDirectionExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					//glslUniform->setUserData( viewDirectionExtractor );
				} 
				else if ( semanticString == "VIEW"  ) 
            {
					ViewMatrixExtractor* viewMatrixExtractor = new ViewMatrixExtractor( glslUniform );
					ss->setAttribute( viewMatrixExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					glslUniform->setUserData( viewMatrixExtractor );
				} 
            else if (  semanticString == "VIEWI" || 
                       semanticString == "VIEWINVERSE" ) 
            {
					ViewMatrixExtractor* viewMatrixExtractor = new ViewMatrixExtractor( glslUniform, ViewMatrixExtractor::INVERSE );
					ss->setAttribute( viewMatrixExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					glslUniform->setUserData( viewMatrixExtractor );
				} 
            else if (  semanticString == "VIEWIT" || 
                       semanticString == "VIEWINVERSETRANSPOSE" ) 
            {
					ViewMatrixExtractor* viewMatrixExtractor = new ViewMatrixExtractor( glslUniform, ViewMatrixExtractor::INVERSETRANSPOSE );
					ss->setAttribute( viewMatrixExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					glslUniform->setUserData( viewMatrixExtractor );
				} 
            else if ( semanticString == "VIEWT" ||
                      semanticString == "VIEWTRANSPOSE" )
            {
					ViewMatrixExtractor* viewMatrixExtractor = new ViewMatrixExtractor( glslUniform, ViewMatrixExtractor::TRANSPOSE) ;
					ss->setAttribute( viewMatrixExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					glslUniform->setUserData( viewMatrixExtractor );
				} 
            else if ( semanticString == "PROJECTION" )
            {
					ViewProjectionExtractor* viewProjectionExtractor = new ViewProjectionExtractor( glslUniform );
					ss->setAttribute( viewProjectionExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					glslUniform->setUserData( viewProjectionExtractor );
				} 
            else if ( semanticString == "PROJECTIONT" ||
                      semanticString == "PROJECTIONTRANSPOSE" )
            {
					ViewProjectionExtractor* viewProjectionExtractor = new ViewProjectionExtractor( glslUniform, ViewProjectionExtractor::TRANSPOSE );
					ss->setAttribute( viewProjectionExtractor, osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON );	
					glslUniform->setUserData( viewProjectionExtractor );
				} 
            else if ( semanticString == "WORLD" ) 
            {
					glslUniform->setUpdateCallback( new worldMatrixCallback() );
				} 
            else if ( semanticString == "WORLDI" || 
                      semanticString == "WORLDINVERSE" ) 
            {
					glslUniform->setUpdateCallback( new worldMatrixInverseCallback() );
				}
            else if ( semanticString == "WORLDT" || 
                      semanticString == "WORLDTRANSPOSE" ) 
            {
					glslUniform->setUpdateCallback( new worldMatrixTransposeCallback() );
				}
            else if ( semanticString == "WORLDIT" || 
                      semanticString == "WORLDINVERSETRANSPOSE" ) 
            {
					glslUniform->setUpdateCallback( new worldMatrixInverseTransposeCallback() );
				}
            else if ( semanticString == "TIMECYCLEPERIOD" )
            {
               // no update is required for this RenderMonkey uniform. It is a constant and happens to be exported with a semantic.
            }
            else
            {
               osg::notify( osg::WARN ) << "Unsupported Semantic: " << semanticString << std::endl;
				   //return true;
            }
			}
		
			if ( glslUniform != NULL ) 
         {
				loadedUniforms[sid] =  glslUniform->getName();
			}                                               
		}

		if ( glslUniform != NULL ) 
      {
			//Check if we just set a sampler* uniform, 
			//update stateset accordingly
			//NOTE: The texture which needs to reside as a uniform
			//was attached to userData before
			if ( glslUniform->getType() == osg::Uniform::SAMPLER_1D ||
			 	  glslUniform->getType() == osg::Uniform::SAMPLER_2D ||
		 		  glslUniform->getType() == osg::Uniform::SAMPLER_2D_SHADOW ||
				  glslUniform->getType() == osg::Uniform::SAMPLER_3D ||
				  glslUniform->getType() == osg::Uniform::SAMPLER_CUBE ) 
			{
				//set the uniform values
				glslUniform->set( texUnitCounter );
				//Add to stateset
				ss->addUniform( glslUniform );
				ss->setTextureAttributeAndModes( texUnitCounter, static_cast<osg::Texture* >( glslUniform->getUserData() ), osg::StateAttribute::ON );
				texUnitCounter++;
			} 
         else 
         {
				//Only add the uniform 
				ss->addUniform( glslUniform );
			}
		}
	}

	//Create effect data, so we know what we've loaded
	EffectData data( ss, loadedUniforms );
	m_loadedEffects[m_pCurrentEffect] = data;
	m_loadedPasses[pDomPass] = data;

	return true;
}
// End AMD change


template <typename T>
bool osgdae::daeReader::processCommonParamTypes(   T              *  paramType, 
                                                   const char     *  symbolName, 
                                                   osg::Uniform   *& newUniformOUT ) 
{
	newUniformOUT = NULL;
	bool isGlslType = false;

	if ( daeSafeCast<domGlsl_param_type>( paramType ) != NULL ) 
   {
		isGlslType = true;
	}

	if ( paramType->getFloat() != NULL ) 
   {
		float data( paramType->getFloat()->getValue() );
		newUniformOUT = new osg::Uniform( symbolName, data );
		return true;
	}

	if ( paramType->getFloat2() != NULL) 
   {
		osg::Vec2f data(
			paramType->getFloat2()->getValue()[0],
			paramType->getFloat2()->getValue()[1]);
		newUniformOUT = new osg::Uniform( symbolName, data );
		return true;
	}

	if ( paramType->getFloat3() != NULL) 
   {
		osg::Vec3f data(
			paramType->getFloat3()->getValue()[0],
			paramType->getFloat3()->getValue()[1],
			paramType->getFloat3()->getValue()[2]);
		newUniformOUT = new osg::Uniform( symbolName, data );
		return true;
	}

	if ( paramType->getFloat4() != NULL ) 
   {
		osg::Vec4f data(
			paramType->getFloat4()->getValue()[0],
			paramType->getFloat4()->getValue()[1],
			paramType->getFloat4()->getValue()[2],
			paramType->getFloat4()->getValue()[3]);

		newUniformOUT = new osg::Uniform( symbolName, data );
		return true;
	}

	if ( paramType->getInt() != NULL ) 
   {
		int data;
		data = paramType->getInt()->getValue();
		newUniformOUT = new osg::Uniform( symbolName, data );
		return true;
	}

	if ( paramType->getInt2() != NULL ) 
   {
		int data[2];
		data[0] = paramType->getInt2()->getValue()[0];
		data[1] = paramType->getInt2()->getValue()[1];
		newUniformOUT = new osg::Uniform( symbolName, data[0], data[1] );
		return true;
	}

	if ( paramType->getInt3() != NULL) 
   {
		int data[3];
		data[0] = paramType->getInt3()->getValue()[0];
		data[1] = paramType->getInt3()->getValue()[1];
		data[2] = paramType->getInt3()->getValue()[2];
		newUniformOUT = new osg::Uniform( symbolName, data[0], data[1], data[2] );
		return true;
	}

	if ( paramType->getInt4() != NULL ) 
   {
		int data[4];
		data[0] = paramType->getInt4()->getValue()[0];
		data[1] = paramType->getInt4()->getValue()[1];
		data[2] = paramType->getInt4()->getValue()[2];
		data[3] = paramType->getInt4()->getValue()[3];
		newUniformOUT = new osg::Uniform( symbolName, data[0], data[1], data[2], data[3] );
		return true;
	}

	if ( paramType->getBool() != NULL ) 
   {
		bool data;
		data = paramType->getBool()->getValue();
		newUniformOUT = new osg::Uniform( symbolName, data );
		return true;
	}

	if ( paramType->getBool2() != NULL ) 
   {
		bool data[2];
		data[0] = paramType->getBool2()->getValue()[0];
		data[1] = paramType->getBool2()->getValue()[1];
		newUniformOUT = new osg::Uniform( symbolName, data[0], data[1] );
		return true;
	}

	if ( paramType->getBool3() != NULL ) 
   {
		bool data[3];
		data[0] = paramType->getBool3()->getValue()[0];
		data[1] = paramType->getBool3()->getValue()[1];
		data[2] = paramType->getBool3()->getValue()[2];
		newUniformOUT = new osg::Uniform( symbolName, data[0], data[1], data[2] );
		return true;
	}

	if ( paramType->getBool4() != NULL ) 
   {
		bool data[4];
		data[0] = paramType->getBool4()->getValue()[0];
		data[1] = paramType->getBool4()->getValue()[1];
		data[2] = paramType->getBool4()->getValue()[2];
		data[3] = paramType->getBool4()->getValue()[3];
		newUniformOUT = new osg::Uniform( symbolName, data[0], data[1], data[2], data[3] );
		return true;
	}

	if ( paramType->getSampler1D() != NULL ) 
   {
		osg::Texture* osgTexture = getTextureFromSurface(std::string(paramType->getSampler1D()->getSource()->getValue()), paramType->getSampler1D());
      if (osgTexture == NULL || !processSamplerCommon1D(osgTexture, paramType->getSampler1D().cast() )) 
      {
			osg::notify( osg::WARN ) << "Converting data from <sampler1D> element failed" << std::endl;
			return false;
		}
		//Create the uniform
		//NOTE: the texture-unit has to be set correctly in the calling method
		//We temporarily save the texture in userData
		newUniformOUT = new osg::Uniform(osg::Uniform::SAMPLER_1D, symbolName);
		newUniformOUT->setUserData(osgTexture);
		return true;
	}

	if ( paramType->getSampler2D() != NULL ) 
   {
      // Here is the point we need to check if the surface has been overridden by an a setaparam 
      // at the technique or material level.
      // We cannot know yet if we are being overridden
		osg::Texture* osgTexture = getTextureFromSurface(  std::string( paramType->getSampler2D()->getSource()->getValue() ), 
                                                         paramType->getSampler2D() );
		
      if ( osgTexture == NULL || !processSampler2D( osgTexture, paramType->getSampler2D().cast() ) ) 
      {
			osg::notify( osg::WARN ) << "Converting data from <sampler2D> element failed" << std::endl;
			return false;
		}
		//Create the uniform
		newUniformOUT = new osg::Uniform( osg::Uniform::SAMPLER_2D, symbolName );
		//NOTE: the texture-unit has to be set correctly in the calling method
		//Attach texture temporarily to userData (as we cannot bind th
		newUniformOUT->setUserData( osgTexture );
		return true;
	}

	if ( paramType->getSamplerRECT() != NULL ) 
   {
		osg::Texture* osgTexture = getTextureFromSurface(std::string(paramType->getSamplerRECT()->getSource()->getValue()), paramType->getSamplerRECT());
		
      if (osgTexture == NULL || !processSampler2D(osgTexture, paramType->getSamplerRECT().cast() ) ) 
      {
			osg::notify( osg::WARN ) << "Converting data from <samplerRECT> element failed" << std::endl;
			return false;
		}
		//Create the uniform
		//NOTE: the texture-unit has to be set correctly in the calling method
		newUniformOUT = new osg::Uniform(osg::Uniform::SAMPLER_2D, symbolName);
		newUniformOUT->setUserData(osgTexture);
		return true;
	}

	if ( paramType->getSamplerDEPTH() != NULL ) 
   {
		osg::Texture* osgTexture = getTextureFromSurface(std::string(paramType->getSamplerDEPTH()->getSource()->getValue()), paramType->getSamplerDEPTH());
		
      if (!processSamplerDEPTH(osgTexture, paramType->getSamplerDEPTH().cast() ) ) 
      {
			osg::notify( osg::WARN ) << "Converting data from <samplerDEPTH> element failed" << std::endl;
			return false;
		}
		//Create the uniform
		//NOTE: the texture-unit has to be set correctly in the calling method
		newUniformOUT = new osg::Uniform(osg::Uniform::SAMPLER_2D_SHADOW, symbolName);
		newUniformOUT->setUserData(osgTexture);
		return true;
	}

	if ( paramType->getSampler3D() != NULL ) 
   {
		osg::Texture* osgTexture = getTextureFromSurface(std::string(paramType->getSampler3D()->getSource()->getValue()), paramType->getSampler3D());
		
      if (osgTexture == NULL || !processSampler3D(osgTexture, paramType->getSampler3D().cast() ) ) 
      {
			osg::notify( osg::WARN ) << "Converting data from <sampler3D> element failed" << std::endl;
			return false;
		}
      //osgTexture->setUseHardwareMipMapGeneration ( false ); // AMD
		//Create the uniform
		//NOTE: the texture-unit has to be set correctly in the calling method
		newUniformOUT = new osg::Uniform(osg::Uniform::SAMPLER_3D, symbolName);
		newUniformOUT->setUserData(osgTexture);
		return true;
	}

	if ( paramType->getSamplerCUBE() != NULL ) 
   {
		osg::Texture* osgTexture = getTextureFromSurface(std::string(  paramType->getSamplerCUBE()->getSource()->getValue()), 
                                                                     paramType->getSamplerCUBE() );
		
      if (osgTexture == NULL || !processSampler3D(osgTexture, paramType->getSamplerCUBE().cast() ) ) 
      {
			osg::notify( osg::WARN ) << "Converting data from <samplerCUBE> element failed" << std::endl;
			return false;
		}
		//Create the uniform
		//NOTE: the texture-unit has to be set correctly in the calling method
		newUniformOUT = new osg::Uniform(osg::Uniform::SAMPLER_CUBE, symbolName);
		newUniformOUT->setUserData(osgTexture);
		return true;
	}

	//NOTE: surface only might be used for rendertargets
	if ( paramType->getSurface() != NULL ) 
   {
      processSurfaceTypes( paramType->getSurface().cast(), std::string("") );
		return true;
	}

	//NOTE: If we load the data from fx_basic_type matrices
	//we have to transpose, as those matrices are in row-major format
	//matrices of glsl_type are expected to be in column-major already
	if ( paramType->getFloat2x2() != NULL ) 
   {
		osg::Matrix2 matrixData;

		if ( isGlslType ) 
      {
			matrixData.set( 
				paramType->getFloat2x2()->getValue()[0], paramType->getFloat2x2()->getValue()[1],
				paramType->getFloat2x2()->getValue()[2], paramType->getFloat2x2()->getValue()[3]
				);
		} 
      else 
      {
			matrixData.set( 
				paramType->getFloat2x2()->getValue()[0], paramType->getFloat2x2()->getValue()[2],
				paramType->getFloat2x2()->getValue()[1], paramType->getFloat2x2()->getValue()[3]
				);
		}

		newUniformOUT = new osg::Uniform(symbolName, matrixData);
		return true;
	}

	if ( paramType->getFloat3x3() != NULL ) 
   {
		osg::Matrix3 matrixData;

		if (isGlslType) 
      {
			matrixData.set( 
				paramType->getFloat3x3()->getValue()[0], paramType->getFloat3x3()->getValue()[1], paramType->getFloat3x3()->getValue()[2],
				paramType->getFloat3x3()->getValue()[3], paramType->getFloat3x3()->getValue()[4], paramType->getFloat3x3()->getValue()[5],
				paramType->getFloat3x3()->getValue()[6], paramType->getFloat3x3()->getValue()[7], paramType->getFloat3x3()->getValue()[8]
				);
		} 
      else 
      {
			matrixData.set( 
				paramType->getFloat3x3()->getValue()[0], paramType->getFloat3x3()->getValue()[3], paramType->getFloat3x3()->getValue()[6],
				paramType->getFloat3x3()->getValue()[1], paramType->getFloat3x3()->getValue()[4], paramType->getFloat3x3()->getValue()[7],
				paramType->getFloat3x3()->getValue()[2], paramType->getFloat3x3()->getValue()[5], paramType->getFloat3x3()->getValue()[8]
				);
		}

		newUniformOUT = new osg::Uniform(symbolName, matrixData);
		return true;
	}

	if ( paramType->getFloat4x4() != NULL ) 
   {
		osg::Matrixf matrixData;

      /*
         printf ( "Matrix data:\n%.1f %.1f %.1f %.1f\n%.1f %.1f %.1f %.1f\n%.1f %.1f %.1f %.1f\n%.1f %.1f %.1f %.1f\n",
               paramType->getFloat4x4()->getValue()[0],
               paramType->getFloat4x4()->getValue()[1],
               paramType->getFloat4x4()->getValue()[2],
               paramType->getFloat4x4()->getValue()[3],
               paramType->getFloat4x4()->getValue()[4],
               paramType->getFloat4x4()->getValue()[5],
               paramType->getFloat4x4()->getValue()[6],
               paramType->getFloat4x4()->getValue()[7],
               paramType->getFloat4x4()->getValue()[8],
               paramType->getFloat4x4()->getValue()[9],
               paramType->getFloat4x4()->getValue()[10],
               paramType->getFloat4x4()->getValue()[11],
               paramType->getFloat4x4()->getValue()[12],
               paramType->getFloat4x4()->getValue()[13],
               paramType->getFloat4x4()->getValue()[14] );
               */

		if (isGlslType) 
      {
			matrixData.set( 
				paramType->getFloat4x4()->getValue()[0], paramType->getFloat4x4()->getValue()[1], paramType->getFloat4x4()->getValue()[2], paramType->getFloat4x4()->getValue()[3], 
				paramType->getFloat4x4()->getValue()[4], paramType->getFloat4x4()->getValue()[5], paramType->getFloat4x4()->getValue()[6], paramType->getFloat4x4()->getValue()[7], 
				paramType->getFloat4x4()->getValue()[8], paramType->getFloat4x4()->getValue()[9], paramType->getFloat4x4()->getValue()[10], paramType->getFloat4x4()->getValue()[11], 
				paramType->getFloat4x4()->getValue()[12], paramType->getFloat4x4()->getValue()[13], paramType->getFloat4x4()->getValue()[14], paramType->getFloat4x4()->getValue()[15] 
				);
		} 
      else 
      {
			matrixData.set(
				paramType->getFloat4x4()->getValue()[0], paramType->getFloat4x4()->getValue()[4], paramType->getFloat4x4()->getValue()[8], paramType->getFloat4x4()->getValue()[12], 
				paramType->getFloat4x4()->getValue()[1], paramType->getFloat4x4()->getValue()[5], paramType->getFloat4x4()->getValue()[9], paramType->getFloat4x4()->getValue()[13], 
				paramType->getFloat4x4()->getValue()[2], paramType->getFloat4x4()->getValue()[6], paramType->getFloat4x4()->getValue()[10], paramType->getFloat4x4()->getValue()[14], 
				paramType->getFloat4x4()->getValue()[3], paramType->getFloat4x4()->getValue()[7], paramType->getFloat4x4()->getValue()[11], paramType->getFloat4x4()->getValue()[15] 
				);
		}

		newUniformOUT = new osg::Uniform(symbolName, matrixData);
		return true;
	}

	osg::notify( osg::WARN ) << "Unknown paramType element" << std::endl;

	return false;
}

//Creates a osg::Uniform out of a glsl-paramtype
//NOTE: returns success of creation, still newUniformOUT IS
//a reference to a pointer. This reference pointer might be NULL
//and the method still might return true. This means, we do not 
//create a uniform out of a glslparamtype, but should not exit because
//of this

bool osgdae::daeReader::processGlslParamType( domGlsl_param_type  *  glslParamType, 
                                             const char           *  symbolName, 
                                             osg::Uniform         *& newUniformOUT ) 
{
	bool success = false;
	success = processCommonParamTypes( glslParamType, symbolName, newUniformOUT );

	return success;
}

bool osgdae::daeReader::processFxCommonParamType(  domFx_basic_type_common *  fxCommonType, 
                                                   const char              *  symbolName, 
                                                   osg::Uniform            *& newUniformOUT ) 
{	
	//NOTE: check for types that are supported
	//in fxCommonProfile but NOT in profileGLSL
	if (fxCommonType->getFloat1x2() != NULL ||
		fxCommonType->getFloat1x3() != NULL ||
		fxCommonType->getFloat1x4() != NULL ||
		fxCommonType->getFloat2x1() != NULL ||
		fxCommonType->getFloat2x3() != NULL ||
		fxCommonType->getFloat2x4() != NULL ||
		fxCommonType->getFloat3x1() != NULL ||
		fxCommonType->getFloat3x2() != NULL ||
		fxCommonType->getFloat3x4() != NULL ||
		fxCommonType->getFloat4x1() != NULL ||
		fxCommonType->getFloat4x2() != NULL ||
		fxCommonType->getFloat4x3() != NULL )
	{
		osg::notify( osg::WARN ) << "Parameter " << symbolName << " is of incompatible type." << std::endl;
		return false;
	}

	bool success = false;
	success = processCommonParamTypes(fxCommonType, symbolName, newUniformOUT);
	return success;
}

// AMD add
domImage* osgdae::daeReader::preCheckDOMImage ( daeIDRef &ref )
{
   domImage* pDomImg = NULL;

	pDomImg = daeSafeCast<domImage >( getElementFromIDRef( ref ) );
	if ( pDomImg == NULL ) 
   {
		osg::notify( osg::WARN ) << "Could not resolve to img element " << ref.getID() << std::endl;
		return NULL;
	}

	//TODO: also load embedded image-data
	if ( pDomImg->getData() ) 
   {
		osg::notify( osg::WARN ) << "Loading embedded image data is not supported yet." << std::endl;
		return NULL;
	}

   return pDomImg;
}

//TODO: There is a lot that has to be done to correctly the surface-types
//However, current COLLADAFX exporters don't always export the right...atm

//Loads the referenced image, apply gl-states, and finally return a osg::Texture
template <typename T>
osg::Texture* osgdae::daeReader::processSurfaceTypes( T* surfaceType, std::string strParamSID ) 
{
	//Load osg::Image out of the init from Array, data in this 
	//array might be used for several surface-types...

	osg::Image* initFromImages[32];
	for ( unsigned int i=0; i<32; i++ ) 
   {
		initFromImages[i] = NULL;
	}
	
	//Set the flip option, if loading DDS
	osgDB::ReaderWriter::Options* loadOptions = new osgDB::ReaderWriter::Options( "dds_flip" );
   unsigned int nImageCount = 0;

   if ( surfaceType->getInit_from_array().getCount() > 0 )
   {
	   for ( unsigned int i=0; i < surfaceType->getInit_from_array().getCount(); i++ ) 
      {
		   //Get the image reference
		   daeIDRef &ref = surfaceType->getInit_from_array()[i]->getValue();

         domImage* pDomImg = preCheckDOMImage ( ref );
         if ( !pDomImg )
         {
            continue;
         }

	      //Check if we already loaded that one
	      ImageMap::iterator it = m_ImageMap.find( pDomImg );
	      if (it != m_ImageMap.end()) 
         {
		      initFromImages[nImageCount] = it->second;		
		      continue;
	      }

		   if ( pDomImg->getInit_from() != NULL ) 
         {
			   //Load the image from the external file
			   std::string imgFileName = getFilenameFromURI( pDomImg->getInit_from()->getValue() );
   		
			   if ( imgFileName.empty() )
            {
				   return NULL;
            }

			   initFromImages[nImageCount] = osgDB::readImageFile( imgFileName, loadOptions );						
			   // Add to map
			   m_ImageMap[ pDomImg ] = initFromImages[nImageCount];

            nImageCount++;
		   }
	   }
   }

// AMD test code
//   else // Check for Init_volume
//   {
//      domFx_surface_init_commonRef rSurfInit = surfaceType->getFx_surface_init_common();
//      if ( rSurfInit )
//      {
//         domFx_surface_init_volume_commonRef rInitVolume = rSurfInit->getInit_volume() ;
//         if ( rInitVolume ) 
//         {
//            xsIDREF &ref = rInitVolume->getAll()->getRef();
//            domImage* pDomImg = preCheckDOMImage ( ref );
//            if ( !pDomImg )
//            {
//               return NULL;
//            }
//			   //Load the image from the external file
//			   std::string imgFileName = getFilenameFromURI( pDomImg->getInit_from()->getValue() );
//			   if ( imgFileName.empty() )
//            {
//				   return NULL;
//            }
//			   initFromImages[nImageCount] = osgDB::readImageFile( imgFileName, loadOptions );						
			   //Add to map
//			   _imageMap[ pDomImg ] = initFromImages[nImageCount];
//         }
//      }
//   }


	// AMD removed as RT is now supported.
        //Check if this surface is rendertarget...
	//if ( surfaceType->getFx_surface_init_common()->getInit_as_target() != NULL ) 
        //{
	//	osg::notify( osg::WARN ) << "Rendertargets are not supported yet." << std::endl;
	//	return NULL;
	//}

	//Check for surface type

	//IMPORTANT: FOR all types check if init_as_null is used...
	//This way no image gets attached yet, but later on, in a <setparam> for 
	//example

	osg::Texture* newTexture = NULL;

	//planar types
	if (surfaceType->getType() == FX_SURFACE_TYPE_ENUM_1D) 
   {
		//Check if we have a init_planar, else use data
		//retrieved from init_from
	}
	
	if ( surfaceType->getType() == FX_SURFACE_TYPE_ENUM_2D ) 
   {
		//Check if we have a init_planar, else use data
		//retrieved from init_from
		if ( surfaceType->getFx_surface_init_common()->getInit_planar() != NULL ) 
      {
			//Load from DDS
			domImage* domImg = NULL;
			domImg = daeSafeCast<domImage >( getElementFromIDRef( surfaceType->getFx_surface_init_common()->getInit_planar()->getAll()->getRef() ));
			if (domImg == NULL) 
         {
				osg::notify( osg::WARN ) << "Could not resolve to img element." << std::endl;
				return NULL;
			}

			osg::Image* img = NULL;

			//Check if we already loaded that one
			ImageMap::iterator it = m_ImageMap.find(domImg);
			if (it != m_ImageMap.end()) 
         {
				img = it->second;		
			} 
         else 
         {
				std::string compoundImgFile = getFilenameFromURI(domImg->getInit_from()->getValue());
				if (compoundImgFile.empty())
					return NULL;

				img = osgDB::readImageFile(compoundImgFile, loadOptions);
				m_ImageMap[domImg] = img;
			}
			newTexture = new osg::Texture2D(img);
		} 
      else if ( initFromImages[0] != NULL ) 
      {
			//TODO: correctly check for mip, slice, face attributes, 
			//Then check if mipmap-generation is enabled/disabled
			//NOTE: No current COLLADAFX-tool is really correctly using this...
			//For now, simply load the first one as the img-base
			newTexture = new osg::Texture2D(initFromImages[0]);
		} 
      else if ( surfaceType->getFx_surface_init_common()->getInit_as_null() != NULL ) 
      {
			//Texture is initialized later
			newTexture = new osg::Texture2D();
		}
      // ADM add
      else if ( surfaceType->getFx_surface_init_common()->getInit_as_target() != NULL ) 
      {
         // This is a Render Target texture
         // Load size
         domInt3 nSize;
         nSize.set3( 128, 128, 0 );
         if ( surfaceType->getSize() )
         {
            // Overwrite the default size if a size element exists.
            nSize = surfaceType->getSize()->getValue();
         }
         else
         {
            osg::notify( osg::WARN ) << "RenderTarget texture has no height and width defined. Uing default values (128x128)" << std::endl;
         }

         // Check to see if there is an RTT target (in this effect/pass) with this name.
         ColladaRTT * pRenderTarget = m_pCurrentColladaEffect->GetRenderTargetByName ( strParamSID );
         if ( !pRenderTarget ) 
         {
            // Check to see if there is an RTT target (from all effects) with this name.
            newTexture = GetRttTexture ( strParamSID ) ;
            if ( newTexture ) 
            {
               return newTexture;
            }
            osg::notify( osg::WARN ) << "Could not locate RenderTarget for " << strParamSID << std::endl;
            return NULL;
         }
         else
         {
            return pRenderTarget->GetTexture();
         }
		}
      // End AMD add
	}

	if (surfaceType->getType() == FX_SURFACE_TYPE_ENUM_RECT)
   {
		//Check if we have a init_planar, else use data
		//retrieved from init_from
	}

	if (surfaceType->getType() == FX_SURFACE_TYPE_ENUM_DEPTH) 
   {
		//Check if we have a init_planar, else use data
		//retrieved from init_from
	}

	if (surfaceType->getType() == FX_SURFACE_TYPE_ENUM_3D) 
   {
      // AMD add
      if ( NULL != surfaceType->getFx_surface_init_common()->getInit_volume() )
      {
			domImage* domImg = NULL;
			domImg = daeSafeCast<domImage >( getElementFromIDRef( surfaceType->getFx_surface_init_common()->getInit_volume()->getAll()->getRef() ));
			if (domImg == NULL) 
         {
				osg::notify( osg::WARN ) << "Could not resolve to img element." << std::endl;
				return NULL;
			}

			osg::Image* img = NULL;

			//Check if we already loaded that one
			ImageMap::iterator it = m_ImageMap.find( domImg );
			if ( it != m_ImageMap.end() ) 
         {
				img = it->second;		
			} 
         else 
         {
				std::string compoundImgFile = getFilenameFromURI( domImg->getInit_from()->getValue() );
				if ( compoundImgFile.empty() )
            {
					return NULL;
            }

				img = osgDB::readImageFile( compoundImgFile, loadOptions );
				m_ImageMap[domImg] = img;
			}
         
         // At this point we have the DDS image
         // We now have to create a 3D texture.
		   osg::Texture3D *new3DTexture = new osg::Texture3D();
         new3DTexture->setImage( img );

         //new3DTexture->setFilter(osg::Texture3D::MIN_FILTER,osg::Texture3D::LINEAR);
         //new3DTexture->setFilter(osg::Texture3D::MAG_FILTER,osg::Texture3D::LINEAR);
         //new3DTexture->setWrap(osg::Texture3D::WRAP_R,osg::Texture3D::REPEAT);
         //new3DTexture->setNumMipmapLevels( 0 );
         //new3DTexture->setUseHardwareMipMapGeneration( false );

         newTexture = new3DTexture;
         // End AMD add
      }
	}
      
	if ( surfaceType->getType() == FX_SURFACE_TYPE_ENUM_CUBE ) 
   {
		//Check if we have a init_cube, else use data
		//retrieved from init_from (but it hat to be correctly used...)
		if ( surfaceType->getFx_surface_init_common()->getInit_cube() != NULL ) 
      {
			//Load all levels from one dds-file
			domFx_surface_init_cube_common* initCube = surfaceType->getFx_surface_init_common()->getInit_cube();

			if ( initCube->getAll() != NULL || initCube->getPrimary() != NULL ) 
         {
				osg::notify( osg::WARN ) << "Surface of type Cube requires initializing via cube-compound images. This is not supported in OSG yet." << std::endl;
				return NULL;
			} 
         else if ( initCube->getFace_array().getCount() > 0 ) 
         {
				//Initialize each face with one compound image 
				//(or simple image, if mipmap_generate is on

				//TODO: Implement constrain, that
				//you have to have all mipmap-levels in compound file
				//if <mipmap_generate> is not true
				//Most COLLADAFX Exporters do not properly take care of this yet

				if ( initCube->getFace_array().getCount() != 6 ) 
            {
					osg::notify( osg::WARN ) << "Face - array in complete" << std::endl;
					return NULL;
				}

				newTexture = new osg::TextureCubeMap();
				osg::TextureCubeMap* newTextureCube = static_cast<osg::TextureCubeMap* >( newTexture );

            // For each face of the cube.
				for ( unsigned int i = 0 ; i < 6 ; i++ ) 
            {
               // Get the DOM image.
					domImage* domImg = NULL;
					domImg = daeSafeCast<domImage >( getElementFromIDRef( initCube->getFace_array()[i]->getRef() ));
					if (domImg == NULL) 
               {
						osg::notify( osg::WARN ) << "Could not resolve to img element." << std::endl;
						return NULL;
					}

               // Check to see if the image has already been loaded.
					osg::Image *img = NULL;
					ImageMap::iterator it = m_ImageMap.find( domImg );
					if ( it != m_ImageMap.end() ) 
               {
						img = it->second;		
					} 
               else 
               {
                  // AMD change + add
                  // Image not previously loaded so go load it now.
						std::string compoundImgFile = getFilenameFromURI( domImg->getInit_from()->getValue() );
						if ( compoundImgFile.empty() )
                  {
							return NULL;
                  }

                  osgDB::ReaderWriter::Options* pLoadOptions = new osgDB::ReaderWriter::Options();

                  // Check for DDS files
                  std::string::size_type nLoc = compoundImgFile.find ( ".dds", 0 );
                  if ( nLoc != std::string::npos )
                  {
                     // Do the DDS flip.
                     pLoadOptions->setOptionString( "dds_flip" );
                  }

                  // Read the image in.
						img = osgDB::readImageFile( compoundImgFile, pLoadOptions );
                  // Store in the image cache.
						m_ImageMap[domImg] = img;
					}

               // Set the index of the face.
					osg::TextureCubeMap::Face osgFace = static_cast<osg::TextureCubeMap::Face>( i );
               // Set the image to be used on the face.
					newTextureCube->setImage( osgFace, img );

               // More work needs to take place here. It is unclear to me how you assign
               // an image to a specific face (e.g. FX_SURFACE_FACE_ENUM_POSITIVE_X) as 
               // the face element here has no face enum.


               switch ( i ) 
               {
               case 0:
                  osgFace = osg::TextureCubeMap::POSITIVE_X;
                  break;
               case 1:
                  osgFace = osg::TextureCubeMap::NEGATIVE_X;
                  break;
               case 2:
                  osgFace = osg::TextureCubeMap::POSITIVE_Y;
                  break;
               case 3:
                  osgFace = osg::TextureCubeMap::NEGATIVE_Y;
                  break;
               case 4:
                  osgFace = osg::TextureCubeMap::POSITIVE_Z;
                  break;
               case 5:
                  osgFace = osg::TextureCubeMap::NEGATIVE_Z;
                  break;
               };
               // End AMD add + change
            }
			}
			//TODO: complete import of cubemaps	
		} 
      else if ( initFromImages[0] != NULL ) 
      {
			//TODO: Import Cubemap from initfrom-2D-components
			//Images are already loaded in initFromImages...(need at least 6 faces, without
			//loading external mipmaps

			//NOTE: there was a bug in the DOM loading dae-files with multiple init_froms, 
			//Andy (SCEA) fixed that bug in SVN 
			
			if ( nImageCount < 6 ) 
         {
				osg::notify( osg::WARN ) << "Not enough image data in init_from Array for Cube Map initialization.";
				return NULL;
			}

			newTexture = new osg::TextureCubeMap();
			osg::TextureCubeMap* newTextureCube = static_cast<osg::TextureCubeMap* >( newTexture );

			for ( unsigned int i=0; i<surfaceType->getInit_from_array().getCount(); i++ ) 
         {
				osg::TextureCubeMap::Face osgFace;

				if (surfaceType->getInit_from_array()[i]->getFace() == FX_SURFACE_FACE_ENUM_POSITIVE_X ) 
            {
					osgFace = osg::TextureCubeMap::POSITIVE_X;
				} 
            else if (surfaceType->getInit_from_array()[i]->getFace() == FX_SURFACE_FACE_ENUM_NEGATIVE_X) 
            {
					osgFace = osg::TextureCubeMap::NEGATIVE_X;
				} 
            else if (surfaceType->getInit_from_array()[i]->getFace() == FX_SURFACE_FACE_ENUM_POSITIVE_Y) 
            {
					osgFace = osg::TextureCubeMap::POSITIVE_Y;
				} 
            else if (surfaceType->getInit_from_array()[i]->getFace() == FX_SURFACE_FACE_ENUM_NEGATIVE_Y) 
            {
					osgFace = osg::TextureCubeMap::NEGATIVE_Y;
				} 
            else if (surfaceType->getInit_from_array()[i]->getFace() == FX_SURFACE_FACE_ENUM_POSITIVE_Z) 
            {
					osgFace = osg::TextureCubeMap::POSITIVE_Z;
				} 
            else if (surfaceType->getInit_from_array()[i]->getFace() == FX_SURFACE_FACE_ENUM_NEGATIVE_Z) 
            {
					osgFace = osg::TextureCubeMap::NEGATIVE_Z;
				}

				newTextureCube->setImage(osgFace, initFromImages[i]);
			}

			//Check if we have a complete Cubemap
			if ( newTextureCube->getImage(osg::TextureCubeMap::POSITIVE_X) == NULL || 
				 newTextureCube->getImage(osg::TextureCubeMap::NEGATIVE_X) == NULL || 
				 newTextureCube->getImage(osg::TextureCubeMap::POSITIVE_Y) == NULL || 
				 newTextureCube->getImage(osg::TextureCubeMap::NEGATIVE_Y) == NULL || 
				 newTextureCube->getImage(osg::TextureCubeMap::POSITIVE_Z) == NULL || 
				 newTextureCube->getImage(osg::TextureCubeMap::NEGATIVE_Z) == NULL ) {
					 osg::notify( osg::WARN ) << "Cubemap initialization incomplete." << std::endl;
					 return NULL;
			}

		} 
      else if (surfaceType->getFx_surface_init_common()->getInit_as_null() != NULL) 
      {
			newTexture = new osg::TextureCubeMap();
		} 
      else 
      {
			osg::notify( osg::WARN ) << "Cubemap: no <init_from>, or <init_cube> found.";
			return NULL;
		}
	}

	if (surfaceType->getType() == FX_SURFACE_TYPE_ENUM_UNTYPED) 
   {
		//Decide behaviour... abort?
	}

	if (surfaceType->getMipmap_generate() != NULL)
   {
		newTexture->setUseHardwareMipMapGeneration( surfaceType->getMipmap_generate()->getValue() );
   }
	
	if (surfaceType->getMip_levels() != NULL) 
   {
		osg::notify( osg::WARN ) << "Handling of max mip-levels not implemented yet" << std::endl;
	}

	//TODO: implement use of generated textures
	domGlsl_surface_type* glslSurfaceType = daeSafeCast<domGlsl_surface_type >( surfaceType );
	if (glslSurfaceType != NULL) 
   {
		if (glslSurfaceType->getGenerator() != NULL) 
      {
			osg::notify( osg::WARN ) << "Handling of <generator>-element is not implemented yet" << std::endl;
			return NULL;
		}
	}
	
	return newTexture;
}

template<typename T>
bool osgdae::daeReader::processSamplerCommon1D(osg::Texture* texture, T* samplerEl) {
	
	//NOTE: We set several defaults which are usually preferred
	//if there's no element set.

	if (samplerEl->getWrap_s() != NULL)
		texture->setWrap( osg::Texture::WRAP_S, m_wrapModeMap[samplerEl->getWrap_s()->getValue()] );
	else
		texture->setWrap( osg::Texture::WRAP_S, osg::Texture::REPEAT );

	if (samplerEl->getMinfilter() != NULL)
		texture->setFilter( osg::Texture::MIN_FILTER, m_filterModeMap[samplerEl->getMinfilter()->getValue()] );
	else
		texture->setFilter( osg::Texture::MIN_FILTER, osg::Texture::LINEAR_MIPMAP_LINEAR );

	if (samplerEl->getMagfilter() != NULL)
		texture->setFilter( osg::Texture::MAG_FILTER, m_filterModeMap[samplerEl->getMagfilter()->getValue()] );
	else
		texture->setFilter( osg::Texture::MIN_FILTER, osg::Texture::LINEAR );

	//TODO: how to map mipfilter??
	if (samplerEl->getMipfilter() != NULL)
		osg::notify( osg::WARN ) << "mipfilter attribute found but ignored."  << std::endl;

	if (samplerEl->getBorder_color() != NULL) 
   {
		const domFloat4& col = samplerEl->getBorder_color()->getValue();
		texture->setBorderColor( osg::Vec4( col[0], col[1], col[2], col[3] ));
	}

	//TODO: implement setting mipmap_maxlevel
	if (samplerEl->getMipmap_maxlevel() != NULL)
		osg::notify( osg::WARN ) << "Attribute mipmap_maxlevel is ignored." << std::endl;

	//TODO: implement setting mipmap_bias
	if (samplerEl->getMipmap_bias() != NULL)
		osg::notify( osg::WARN ) << "Attribute mipmap_bias is ignored." << std::endl;
	
	return true;
}

template <typename T> 
bool osgdae::daeReader::processSampler2D(osg::Texture* texture, T* samplerEl) {
	if (samplerEl->getWrap_t() != NULL)
		texture->setWrap( osg::Texture::WRAP_T, m_wrapModeMap[samplerEl->getWrap_t()->getValue()] );
	else
		texture->setWrap( osg::Texture::WRAP_T, osg::Texture::REPEAT );

	return ( processSamplerCommon1D(texture, samplerEl) );
}

template <typename T>
bool osgdae::daeReader::processSampler3D(osg::Texture* texture, T* samplerEl) 
{
	if ( samplerEl->getWrap_p() != NULL )
   {
		texture->setWrap( osg::Texture::WRAP_R, m_wrapModeMap[samplerEl->getWrap_p()->getValue()] );
   }
   else
   {
      texture->setWrap( osg::Texture::WRAP_R, osg::Texture::REPEAT );
   }

	return ( processSampler2D( texture, samplerEl ));
}

template <typename T>
bool osgdae::daeReader::processSamplerDEPTH(osg::Texture* texture, T* samplerEl) {
	if (samplerEl->getWrap_s() != NULL)
		texture->setWrap( osg::Texture::WRAP_S, m_wrapModeMap[samplerEl->getWrap_s()->getValue()] );
	else
		texture->setWrap( osg::Texture::WRAP_S, osg::Texture::REPEAT );

	if (samplerEl->getWrap_t() != NULL)
		texture->setWrap( osg::Texture::WRAP_T, m_wrapModeMap[samplerEl->getWrap_t()->getValue()] );
	else
		texture->setWrap( osg::Texture::WRAP_T, osg::Texture::REPEAT );
	
	if (samplerEl->getMinfilter() != NULL)
		texture->setFilter( osg::Texture::MIN_FILTER, m_filterModeMap[samplerEl->getMinfilter()->getValue()] );
	else
		texture->setFilter( osg::Texture::MIN_FILTER, osg::Texture::LINEAR_MIPMAP_LINEAR );

	if (samplerEl->getMagfilter() != NULL)
		texture->setFilter( osg::Texture::MAG_FILTER, m_filterModeMap[samplerEl->getMagfilter()->getValue()] );
	else
		texture->setFilter( osg::Texture::MAG_FILTER, osg::Texture::LINEAR );

	return true;
}


// AMD add
domProfile_GLSL * osgdae::daeReader::GetGLSLProfile ( const domEffect *pCurrentEffect ) const
{
	domProfile_GLSL* profileGLSL = NULL;
	for ( unsigned int i = 0 ; i < pCurrentEffect->getFx_profile_abstract_array().getCount(); i++ ) 
   {
		profileGLSL = daeSafeCast<domProfile_GLSL >( pCurrentEffect->getFx_profile_abstract_array()[i] );
		if ( profileGLSL != NULL )
			break;
	}
   if ( profileGLSL == NULL ) 
   {
		osg::notify( osg::WARN ) << "Error getting profileGLSL while processing TextureFromSurface" << std::endl;
		return NULL;
	}

   return profileGLSL;
}

domGlsl_surface_type * osgdae::daeReader::getSurfaceBySid ( const std::string& sid ) const
{
   std::string source = std::string("./") + sid;

   domProfile_GLSL * pProfileGLSL = GetGLSLProfile ( m_pCurrentEffect );
   if ( !pProfileGLSL ) 
   {
		return NULL;
	}

	daeElement* resolvedElement = NULL;
	daeElement* container = pProfileGLSL;

	// At first try to resolve in profile-scope only.
	daeSIDResolver surfaceResolveProfile( pProfileGLSL, source.c_str() );
	resolvedElement = surfaceResolveProfile.getElement();
   // Carry on up to the effect level. 
	if ( resolvedElement == NULL ) 
   {
		do 
      {
			container = container->getParentElement();
			daeSIDResolver surfaceResolve( container, source.c_str() );
			resolvedElement = surfaceResolve.getElement();
			if ( resolvedElement != NULL )
         {
				break;
         }
		} 
      while( daeSafeCast<domEffect >( container ) != m_pCurrentEffect );
	}

   // Cast to a newparam
	domGlsl_newparam * glslNewparam = daeSafeCast<domGlsl_newparam >( resolvedElement );
   if ( !glslNewparam )
   {
      return NULL;
   }

   domGlsl_surface_type * pSurfaceType = NULL;

   // Check the type
   domGlsl_param_typeRef  rType = glslNewparam->getGlsl_param_type();
   if ( rType && rType->getSurface() != NULL ) 
   {
      pSurfaceType = glslNewparam->getGlsl_param_type()->getSurface();
   }

   return pSurfaceType;
}
// End AMD add

// AMD change
osg::Texture* osgdae::daeReader::getTextureFromSurface( const std::string& surfRef, daeElement* samplerEl ) 
{
	// Check to see if the render target surface has already been created.
   // It should in theory. A previous pass must have created the RTT for it
   // to be bound in a subsequent pass.
   osg::Texture *pRTT_Texture = NULL ;
   // Check to see if the texture is an RTT target.
   // RTT are rendered into and do not need intializing with a texture.
   RTTTextureMap::iterator it = m_RTTTextureMap.find( surfRef ) ;
   if ( it != m_RTTTextureMap.end() )
   {
      pRTT_Texture = it->second ;
   }
   if ( pRTT_Texture ) 
   {
      return pRTT_Texture;
   }
   // End AND change
   
   // Resolve to surface newparam
	// search in next upper level until effect-scope is reached
	std::string source = std::string("./") + surfRef;
	daeElement* container = samplerEl->getParentElement();
	daeElement* resolvedElement = NULL;

	// Be sure to resolve to profile_GLSL only
	// NOTE: this should be settable via daeSIDResolver->setProfile(..)
	// but not working yet, so workaround is provided
   domProfile_GLSL * profileGLSL = GetGLSLProfile ( m_pCurrentEffect );
   if ( !profileGLSL ) 
   {
		return NULL;
	}

	//At first try to resolve in profile-scope only
	daeSIDResolver surfaceResolveProfile( profileGLSL, source.c_str() );
	resolvedElement = surfaceResolveProfile.getElement();
	if ( resolvedElement == NULL ) 
   {
		do 
      {
			container = container->getParentElement();
			daeSIDResolver surfaceResolve( container, source.c_str() );
			resolvedElement = surfaceResolve.getElement();
			if ( resolvedElement != NULL )
         {
				break;
         }
		} 
      while( daeSafeCast<domEffect >( container ) != m_pCurrentEffect );
	}

	domGlsl_newparam        * glslNewparam    = daeSafeCast<domGlsl_newparam >( resolvedElement );
	domFx_newparam_common   * commonNewparam  = daeSafeCast<domFx_newparam_common >( resolvedElement );

   // AMD add
   // Need to look for the surface in material setparams to see if the glslParam has been overridden.
   unsigned int nMaterialParamIndex  = 0 ;
   bool bOveriddenByMaterial  = OveriddenByMaterial ( surfRef, m_pCurrentDomMaterial,    nMaterialParamIndex ) ;
   if ( glslNewparam && bOveriddenByMaterial )
   {
      domInstance_effectRef rInstanceEffect = m_pCurrentDomMaterial->getInstance_effect();
      if ( rInstanceEffect == NULL ) 
      {
         osg::notify( osg::WARN ) << "Error getting <instance_effect> from <material>-element. " << std::endl;
         return false;
      }
      domFx_basic_type_commonRef  rType = rInstanceEffect->getSetparam_array()[nMaterialParamIndex]->getFx_basic_type_common();
		if ( rType && rType->getSurface() != NULL ) 
      {
			osg::Texture* referencedTexture = processSurfaceTypes(  rType->getSurface().cast(), glslNewparam->getSid() );
			m_loadedSurfacesMap[resolvedElement] = referencedTexture;
			return referencedTexture;
		}
   }
   else // Not overridden
   {
      // End AMD add
      if ( glslNewparam != NULL ) 
      {
		   // we loose information like semantic, or annotations 
		   // by directly creating runtime-information

		   //Get the surface element
		   if ( glslNewparam->getGlsl_param_type()->getSurface() != NULL ) 
         {
			   osg::Texture* referencedTexture = processSurfaceTypes(   glslNewparam->getGlsl_param_type()->getSurface().cast(), glslNewparam->getSid() );
			   m_loadedSurfacesMap[resolvedElement] = referencedTexture;
			   return referencedTexture;
		   } 
         else 
         {
			   osg::notify( osg::WARN ) << "Referenced <newparam> " << source << " is NOT a surface." << std::endl;
			   return NULL;
		   }
	   } 
      else if ( commonNewparam != NULL ) 
      {
		   if ( commonNewparam->getFx_basic_type_common()->getSurface() != NULL ) 
         {
			   osg::Texture* referencedTexture = processSurfaceTypes(   commonNewparam->getFx_basic_type_common()->getSurface().cast(), glslNewparam->getSid());
			   //Save to map
			   m_loadedSurfacesMap[resolvedElement] = referencedTexture;
			   return referencedTexture;
		   } 
         else 
         {
			   osg::notify( osg::WARN ) << "Referenced <newparam> " << source << " is NOT a surface." << std::endl;
			   return NULL;
		   }
	   } 
      else 
      {
		   osg::notify( osg::WARN ) << "Unknown type of <newparam> " << std::endl;
		   return NULL;
	   }
   }

	return NULL;
}

// AMD add
bool osgdae::daeReader::GetStateSet(   domProfile_GLSL::domTechnique::domPass *pDomPass,
                                       osg::StateSet                          *& out_pStateSet,
                                       osg::Program                           *& out_pProgram  )
{
	//Get a cached effect, and update ONLY bind-array
	//we would need this for correct world-matrices, etc...
	PassGLSLMap::iterator it = m_passGLSLMap.find( pDomPass );
	if ( it != m_passGLSLMap.end() ) 
   {
		//Already processed this profile, only update the bind-array
		//with per-object callbacks...
		//Copy the stateset (ss->merge does NOT work with uniform updates...)
		out_pStateSet = new osg::StateSet( *(it->second) );
		return  true;
	} 
 
	//Create Stateset return for this effect
	out_pStateSet = new osg::StateSet();
	out_pProgram = new osg::Program();

   return false;
}

osg::Shader* osgdae::daeReader::LoadShader ( daeElement *pResolvedElement, domGlsl_pipeline_stage eStage )
{
   std::ostringstream shaderSrcOss;

   // ShaderSrc could be in <code> or <include>
   domFx_include_common* srcIncludeEl = daeSafeCast<domFx_include_common >( pResolvedElement );
   if ( srcIncludeEl != NULL ) 
   {
      //Load from external file
      srcIncludeEl->getUrl().validate();

      char* path = NULL;
      char* filename = NULL;
      if ( std::string( srcIncludeEl->getUrl().getProtocol() ) == std::string( "file" ) )
      {
	      size_t bufSize = 1; //for the null char
	      if ( srcIncludeEl->getUrl().getFilepath() != NULL )
	      {
		      bufSize += strlen( srcIncludeEl->getUrl().getFilepath() );
	      }
	      if ( srcIncludeEl->getUrl().getFile() != NULL )
	      {
		      bufSize += strlen( srcIncludeEl->getUrl().getFile() );
	      }
	      path = new char[bufSize+1];
	      if ( !srcIncludeEl->getUrl().getPath( path, daeInt(bufSize) ) )
	      {
		      osg::notify( osg::WARN ) << "Unable to get Path from URI" << std::endl;
		      return NULL;
	      }

   #ifdef WIN32
	      // under windows the URI passed back from the dom includes a / i.e. /c:/path/file, so we need to jump over the leading /
	      filename = path+1;
   #else
	      filename = path;
   #endif
      } 
      else 
      {
	      osg::notify( osg::WARN ) << "Wrong Protocol for ShaderSrc-Include" << std::endl;
	      return NULL;
      }

      //Read the file
      std::ifstream ifs;
      ifs.open( filename, std::ifstream::in );
      if (ifs.fail()) 
      {
	      osg::notify( osg::WARN ) << "File " << filename << " does not exist." << std::endl;
	      return NULL;
      }

      shaderSrcOss << ifs.rdbuf();
      ifs.close();

      //Delete the char-buffer
      if ( path != NULL )
      {
	      delete[] path;
      }
   }

   // The code is in the COLLADA doc (not an include file).
   domFx_code_profile* srcCodeEl = daeSafeCast<domFx_code_profile >( pResolvedElement );
   if ( srcCodeEl != NULL ) 
   {
      shaderSrcOss << srcCodeEl->getValue();
   }

   osg::Shader::Type osgShaderType;

   if ( eStage == GLSL_PIPELINE_STAGE_FRAGMENTPROGRAM )
   {
      osgShaderType = osg::Shader::FRAGMENT;
   }
   else if ( eStage == GLSL_PIPELINE_STAGE_VERTEXPROGRAM )
   {
      osgShaderType = osg::Shader::VERTEX;
   }
   else 
   {
      osg::notify( osg::WARN ) << "Unknown Shader stage." << std::endl;
      return NULL;
   }

   //Create the shader
   osg::Shader* pShader = new osg::Shader( osgShaderType, shaderSrcOss.str() );

   return pShader;
}

daeElement* osgdae::daeReader::ResolveElement ( std::string &strSource,
                                                daeElement *pParent )
{
   //Resolve in steps to technique->profile->effect scope
   daeElement* container = pParent;
   daeElement* resolvedElement;
   for ( int j=0; j<3; j++ ) 
   {
      daeSIDResolver resolver( container, strSource.c_str() );
      resolvedElement = resolver.getElement();
      //If we could resolve break, else try with parent-scope
      if ( resolvedElement != NULL )
      {
	      break;
      }
      container = container->getParentElement();
   }

   if ( container == NULL ) 
   {
      osg::notify( osg::WARN ) << "Could not resolve GLSL-Shader-Src." << std::endl;
      return NULL;
   }

   return resolvedElement ;
}

void osgdae::daeReader::AddVertexAttributes ( 	osg::Program   *sProgram, 
                                                domInstance_material::domBind_vertex_input_Array & rBindVertexInputArray )
{
   // Add the generic vertex attribute bindings.
   size_t nBviCount = rBindVertexInputArray.getCount();
   for ( size_t nBviIndex = 0; nBviIndex < nBviCount; nBviIndex++ )
   {
      domInstance_material::domBind_vertex_inputRef bviRef = rBindVertexInputArray.get( nBviIndex );

      xsNCName strBindToName      = bviRef->getSemantic();
      xsNCName strInputStreamName = bviRef->getInput_semantic();
      domUint  nInputSet          = bviRef->getInput_set();

      if ( strcmp ( strInputStreamName, "TANGENT" ) == 0 )
      {
         nInputSet = TANGENT_TEX_UNIT ;
      }
      else if ( strcmp ( strInputStreamName, "BITANGENT" ) == 0 )
      {
         nInputSet = BITANGENT_TEX_UNIT ;
      }
      else if ( strcmp ( strInputStreamName, "BINORMAL" ) == 0 )
      {
         nInputSet = BITANGENT_TEX_UNIT ;
      }

      sProgram->addBindAttribLocation ( strBindToName, nInputSet );
      //sProgram->addBindAttribLocation ( strInputStreamName, nInputSet );
   }
}

osg::Texture * osgdae::daeReader::GetRttTexture ( std::string strParamSID ) 
{
   RTTTextureMap::iterator it = m_RTTTextureMap.find( strParamSID ) ;
   
   if ( it != m_RTTTextureMap.end() )
   {
      return it->second ;
   }

   return NULL;
}

void osgdae::daeReader::HandleEffectExtras ( domEffect * pDomEffect, ColladaEffectNode * pEffect )
{
   std::string strValue ;

   if ( daeReader::FindInExtra( pDomEffect, 
                     "RenderMonkey",
                     "RenderMonkey_TimeCycle",
                     "",
                     strValue ) )
   {

      float fTimeCycle = fDefaultRenderMonkeyTimeCycle;
      sscanf ( strValue.c_str(), "%f", &fTimeCycle );
      pEffect->SetTimeCycle ( fTimeCycle );
   }
}// End AMD add
